# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: qm/pb/frontend.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'qm/pb/frontend.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from qm.grpc.qm.pb import inc_qua_pb2 as qm_dot_pb_dot_inc__qua__pb2
from qm.grpc.qm.pb import inc_qua_config_pb2 as qm_dot_pb_dot_inc__qua__config__pb2
from qm.grpc.qm.pb import inc_qm_api_pb2 as qm_dot_pb_dot_inc__qm__api__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2
from qm.grpc.qm.pb import compiler_pb2 as qm_dot_pb_dot_compiler__pb2
from qm.grpc.qm.pb import job_results_pb2 as qm_dot_pb_dot_job__results__pb2
from qm.grpc.qm.pb import qm_manager_pb2 as qm_dot_pb_dot_qm__manager__pb2
from qm.grpc.qm.pb import general_messages_pb2 as qm_dot_pb_dot_general__messages__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14qm/pb/frontend.proto\x12\x10qm.grpc.frontend\x1a\x13qm/pb/inc_qua.proto\x1a\x1aqm/pb/inc_qua_config.proto\x1a\x16qm/pb/inc_qm_api.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1egoogle/protobuf/wrappers.proto\x1a\x14qm/pb/compiler.proto\x1a\x17qm/pb/job_results.proto\x1a\x16qm/pb/qm_manager.proto\x1a\x1cqm/pb/general_messages.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"k\n\rQueuePosition\x12%\n\x03\x65nd\x18\x01 \x01(\x0b\x32\x16.google.protobuf.EmptyH\x00\x12\'\n\x05start\x18\x02 \x01(\x0b\x32\x16.google.protobuf.EmptyH\x00\x42\n\n\x08position\"\x98\x01\n\x11\x41\x64\x64ToQueueRequest\x12\x18\n\x10quantumMachineId\x18\x01 \x01(\t\x12\x31\n\x10highLevelProgram\x18\x02 \x01(\x0b\x32\x17.qm.grpc.qua.QuaProgram\x12\x36\n\rqueuePosition\x18\x03 \x01(\x0b\x32\x1f.qm.grpc.frontend.QueuePosition\"d\n\x12\x41\x64\x64ToQueueResponse\x12\n\n\x02ok\x18\x01 \x01(\x08\x12\r\n\x05jobId\x18\x02 \x01(\t\x12\x33\n\x08messages\x18\x03 \x03(\x0b\x32!.qm.grpc.compiler.CompilerMessage\"#\n\x10WaveformOverride\x12\x0f\n\x07samples\x18\x01 \x03(\x01\"\xb2\x01\n\x12\x45xecutionOverrides\x12\x46\n\twaveforms\x18\x01 \x03(\x0b\x32\x33.qm.grpc.frontend.ExecutionOverrides.WaveformsEntry\x1aT\n\x0eWaveformsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x31\n\x05value\x18\x02 \x01(\x0b\x32\".qm.grpc.frontend.WaveformOverride:\x02\x38\x01\"\xc2\x01\n\x19\x41\x64\x64\x43ompiledToQueueRequest\x12\x18\n\x10quantumMachineId\x18\x01 \x01(\t\x12\x11\n\tprogramId\x18\x02 \x01(\t\x12\x36\n\rqueuePosition\x18\x03 \x01(\x0b\x32\x1f.qm.grpc.frontend.QueuePosition\x12@\n\x12\x65xecutionOverrides\x18\x04 \x01(\x0b\x32$.qm.grpc.frontend.ExecutionOverrides\"o\n\x1a\x41\x64\x64\x43ompiledToQueueResponse\x12\n\n\x02ok\x18\x01 \x01(\x08\x12\r\n\x05jobId\x18\x02 \x01(\t\x12\x36\n\x06\x65rrors\x18\x03 \x03(\x0b\x32&.qm.grpc.general_messages.ErrorMessage\"]\n\x0e\x43ompileRequest\x12\x18\n\x10quantumMachineId\x18\x01 \x01(\t\x12\x31\n\x10highLevelProgram\x18\x02 \x01(\x0b\x32\x17.qm.grpc.qua.QuaProgram\"e\n\x0f\x43ompileResponse\x12\n\n\x02ok\x18\x01 \x01(\x08\x12\x11\n\tprogramId\x18\x02 \x01(\t\x12\x33\n\x08messages\x18\x03 \x03(\x0b\x32!.qm.grpc.compiler.CompilerMessage\"G\n\x1cGetJobExecutionStatusRequest\x12\x18\n\x10quantumMachineId\x18\x01 \x01(\t\x12\r\n\x05jobId\x18\x02 \x01(\t\"U\n\x1dGetJobExecutionStatusResponse\x12\x34\n\x06status\x18\x01 \x01(\x0b\x32$.qm.grpc.frontend.JobExecutionStatus\"\xe0\n\n\x12JobExecutionStatus\x12?\n\x07unknown\x18\x01 \x01(\x0b\x32,.qm.grpc.frontend.JobExecutionStatus.UnknownH\x00\x12?\n\x07pending\x18\x02 \x01(\x0b\x32,.qm.grpc.frontend.JobExecutionStatus.PendingH\x00\x12?\n\x07running\x18\x03 \x01(\x0b\x32,.qm.grpc.frontend.JobExecutionStatus.RunningH\x00\x12\x43\n\tcompleted\x18\x04 \x01(\x0b\x32..qm.grpc.frontend.JobExecutionStatus.CompletedH\x00\x12\x41\n\x08\x63\x61nceled\x18\x05 \x01(\x0b\x32-.qm.grpc.frontend.JobExecutionStatus.CanceledH\x00\x12?\n\x07loading\x18\x06 \x01(\x0b\x32,.qm.grpc.frontend.JobExecutionStatus.LoadingH\x00\x12;\n\x05\x65rror\x18\x07 \x01(\x0b\x32*.qm.grpc.frontend.JobExecutionStatus.ErrorH\x00\x12\x45\n\nprocessing\x18\x08 \x01(\x0b\x32/.qm.grpc.frontend.JobExecutionStatus.ProcessingH\x00\x1a\t\n\x07Unknown\x1a\x62\n\x07Pending\x12\x17\n\x0fpositionInQueue\x18\x01 \x01(\x05\x12-\n\ttimeAdded\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0f\n\x07\x61\x64\x64\x65\x64\x42y\x18\x03 \x01(\t\x1az\n\x07Running\x12-\n\ttimeAdded\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0f\n\x07\x61\x64\x64\x65\x64\x42y\x18\x02 \x01(\t\x12/\n\x0btimeStarted\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a\xaf\x01\n\tCompleted\x12-\n\ttimeAdded\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0f\n\x07\x61\x64\x64\x65\x64\x42y\x18\x02 \x01(\t\x12/\n\x0btimeStarted\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\rtimeCompleted\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a|\n\x08\x43\x61nceled\x12-\n\ttimeAdded\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0f\n\x07\x61\x64\x64\x65\x64\x42y\x18\x02 \x01(\t\x12\x30\n\x0ctimeCanceled\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1aI\n\x07Loading\x12-\n\ttimeAdded\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0f\n\x07\x61\x64\x64\x65\x64\x42y\x18\x02 \x01(\t\x1aK\n\x05\x45rror\x12\x31\n\rerrorMessages\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.ListValue\x12\x0f\n\x07\x61\x64\x64\x65\x64\x42y\x18\x02 \x01(\t\x1a}\n\nProcessing\x12-\n\ttimeAdded\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0f\n\x07\x61\x64\x64\x65\x64\x42y\x18\x02 \x01(\t\x12/\n\x0btimeStarted\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x08\n\x06status\"\xca\x01\n\x16GetPendingJobsResponse\x12N\n\x0bpendingJobs\x18\x01 \x03(\x0b\x32\x39.qm.grpc.frontend.GetPendingJobsResponse.PendingJobsEntry\x1a`\n\x10PendingJobsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12;\n\x05value\x18\x02 \x01(\x0b\x32,.qm.grpc.frontend.JobExecutionStatus.Pending:\x02\x38\x01\"\xc3\x01\n\x0eJobQueryParams\x12\x18\n\x10quantumMachineId\x18\x01 \x01(\t\x12\x32\n\x05jobId\x18\x02 \x01(\x0b\x32#.qm.grpc.frontend.QueryValueMatcher\x12\x33\n\x06userId\x18\x03 \x01(\x0b\x32#.qm.grpc.frontend.QueryValueMatcher\x12.\n\x08position\x18\x04 \x01(\x0b\x32\x1c.google.protobuf.UInt32Value\"@\n\x11QueryValueMatcher\x12\r\n\x03\x61ny\x18\x01 \x01(\x08H\x00\x12\x0f\n\x05value\x18\x02 \x01(\tH\x00\x42\x0b\n\tvalueType\"9\n\x19RemovePendingJobsResponse\x12\x1c\n\x14numbersOfJobsRemoved\x18\x01 \x01(\x05\"\xf9\x01\n\x11SimulationRequest\x12-\n\x06\x63onfig\x18\x01 \x01(\x0b\x32\x1d.qm.grpc.qua_config.QuaConfig\x12\x31\n\x10highLevelProgram\x18\x02 \x01(\x0b\x32\x17.qm.grpc.qua.QuaProgram\x12=\n\x08simulate\x18\x03 \x01(\x0b\x32+.qm.grpc.frontend.ExecutionRequest.Simulate\x12\x43\n\x15\x63ontrollerConnections\x18\x06 \x03(\x0b\x32$.qm.grpc.frontend.InterOpxConnection\"3\n\x0fInterOpxAddress\x12\x12\n\ncontroller\x18\x01 \x01(\t\x12\x0c\n\x04left\x18\x02 \x01(\x08\"O\n\x0eInterOpxTarget\x12\x33\n\x06\x64irect\x18\x01 \x01(\x0b\x32!.qm.grpc.frontend.InterOpxAddressH\x00\x42\x08\n\x06target\"<\n\x0fInterOpxChannel\x12\x12\n\ncontroller\x18\x01 \x01(\t\x12\x15\n\rchannelNumber\x18\x02 \x01(\x05\"\xad\x04\n\x12InterOpxConnection\x12\x35\n\x06source\x18\x01 \x01(\x0b\x32!.qm.grpc.frontend.InterOpxAddressB\x02\x18\x01\x12\x34\n\x06target\x18\x02 \x01(\x0b\x32 .qm.grpc.frontend.InterOpxTargetB\x02\x18\x01\x12Q\n\x10\x61\x64\x64ressToAddress\x18\x03 \x01(\x0b\x32\x35.qm.grpc.frontend.InterOpxConnection.AddressToAddressH\x00\x12Q\n\x10\x63hannelToChannel\x18\x04 \x01(\x0b\x32\x35.qm.grpc.frontend.InterOpxConnection.ChannelToChannelH\x00\x1ax\n\x10\x41\x64\x64ressToAddress\x12\x31\n\x06source\x18\x01 \x01(\x0b\x32!.qm.grpc.frontend.InterOpxAddress\x12\x31\n\x06target\x18\x02 \x01(\x0b\x32!.qm.grpc.frontend.InterOpxAddress\x1ax\n\x10\x43hannelToChannel\x12\x31\n\x06source\x18\x01 \x01(\x0b\x32!.qm.grpc.frontend.InterOpxChannel\x12\x31\n\x06target\x18\x02 \x01(\x0b\x32!.qm.grpc.frontend.InterOpxChannelB\x10\n\x0e\x63onnectionType\"\xc3\x02\n\x12SimulationResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\r\n\x05jobId\x18\x02 \x01(\t\x12K\n\x16\x63onfigValidationErrors\x18\x03 \x03(\x0b\x32+.qm.grpc.qm_manager.ConfigValidationMessage\x12O\n\x18physicalValidationErrors\x18\x04 \x03(\x0b\x32-.qm.grpc.qm_manager.PhysicalValidationMessage\x12\x33\n\x08messages\x18\x05 \x03(\x0b\x32!.qm.grpc.compiler.CompilerMessage\x12:\n\tsimulated\x18\x06 \x01(\x0b\x32\'.qm.grpc.frontend.SimulatedResponsePart\"$\n\x13IsJobRunningRequest\x12\r\n\x05jobId\x18\x01 \x01(\t\"8\n\x14IsJobRunningResponse\x12\r\n\x05jobId\x18\x01 \x01(\t\x12\x11\n\tisRunning\x18\x02 \x01(\x08\"*\n\x19IsJobAcquiringDataRequest\x12\r\n\x05jobId\x18\x01 \x01(\t\"\xdb\x01\n\x1aIsJobAcquiringDataResponse\x12\r\n\x05jobId\x18\x01 \x01(\t\x12U\n\x0f\x61\x63quiringStatus\x18\x02 \x01(\x0e\x32<.qm.grpc.frontend.IsJobAcquiringDataResponse.AcquiringStatus\"W\n\x0f\x41\x63quiringStatus\x12\x13\n\x0f\x41\x43QUIRE_STOPPED\x10\x00\x12\x16\n\x12NO_DATA_TO_ACQUIRE\x10\x01\x12\x17\n\x13HAS_DATA_TO_ACQUIRE\x10\x02\"K\n\rQmDataRequest\x12:\n\x10io_value_Request\x18\x01 \x03(\x0b\x32 .qm.grpc.frontend.IOValueRequest\"\x99\x02\n\x0eQmDataResponse\x12K\n\x11io_value_response\x18\x01 \x03(\x0b\x32\x30.qm.grpc.frontend.QmDataResponse.IOValueResponse\x12\x0f\n\x07success\x18\x02 \x01(\x08\x12\x36\n\x06\x65rrors\x18\x03 \x03(\x0b\x32&.qm.grpc.general_messages.ErrorMessage\x1aq\n\x0fIOValueResponse\x12\x31\n\x07request\x18\x01 \x01(\x0b\x32 .qm.grpc.frontend.IOValueRequest\x12+\n\x06values\x18\x02 \x01(\x0b\x32\x1b.qm.grpc.compiler.QuaValues\"\x80\x01\n\x0eIOValueRequest\x12\x19\n\rpulser_number\x18\x01 \x01(\rB\x02\x18\x01\x12\x11\n\tio_number\x18\x02 \x01(\r\x12\x0f\n\x05jobId\x18\x03 \x01(\tH\x00\x12\x1a\n\x10quantumMachineId\x18\x04 \x01(\tH\x00\x42\x13\n\x11identifier_one_of\"9\n\rResumeRequest\x12\r\n\x05jobId\x18\x01 \x01(\t\x12\x19\n\rpulsersBitmap\x18\x02 \x01(\rB\x02\x18\x01\"\x10\n\x0eResumeResponse\"$\n\x13PausedStatusRequest\x12\r\n\x05jobId\x18\x01 \x01(\t\"O\n\x14PausedStatusResponse\x12\x19\n\rpulsersBitmap\x18\x01 \x01(\rB\x02\x18\x01\x12\n\n\x02ok\x18\x02 \x01(\x08\x12\x10\n\x08isPaused\x18\x03 \x01(\x08\"\"\n\x11PullResultRequest\x12\r\n\x05jobId\x18\x01 \x01(\t\"=\n\x1aPullAnalysedResultsRequest\x12\r\n\x05jobId\x18\x01 \x01(\t\x12\x10\n\x08metadata\x18\x02 \x01(\t\"\x94\x01\n\x12PullResultResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\x0e\n\x06offset\x18\x02 \x01(\r\x12\x11\n\tbytesLeft\x18\x03 \x01(\r\x12\x16\n\x0e\x63ontrollerName\x18\x04 \x01(\t\x12\x19\n\x11\x61pplicationNumber\x18\x05 \x01(\r\x12\x1a\n\x12implementationName\x18\x06 \x01(\t\"b\n\x13HealthCheckResponse\x12\x0f\n\x07message\x18\x01 \x03(\t\x12\n\n\x02ok\x18\x02 \x01(\x08\x12\x15\n\rerrorMessages\x18\x03 \x03(\t\x12\x17\n\x0fwarningMessages\x18\x04 \x03(\t\"\xec\x0b\n\x10\x45xecutionRequest\x12\x33\n\x10highLevelProgram\x18\x01 \x01(\x0b\x32\x17.qm.grpc.qua.QuaProgramH\x00\x12\x1d\n\x0flowLevelProgram\x18\x02 \x01(\x0c\x42\x02\x18\x01H\x00\x12\x1b\n\x13streamDurationLimit\x18\x03 \x01(\r\x12\x17\n\x0fstreamDataLimit\x18\x04 \x01(\r\x12\x16\n\x0e\x66orceExecution\x18\x05 \x01(\x08\x12=\n\x08simulate\x18\x06 \x01(\x0b\x32+.qm.grpc.frontend.ExecutionRequest.Simulate\x12\x0e\n\x06\x64ryRun\x18\x07 \x01(\x08\x12\x18\n\x10quantumMachineId\x18\x08 \x01(\t\x1a\xba\t\n\x08Simulate\x12\x10\n\x08\x64uration\x18\x01 \x01(\r\x12!\n\x15simulateAnalogOutputs\x18\x02 \x01(\x08\x42\x02\x18\x01\x12\"\n\x16simulateDigitalOutputs\x18\x03 \x01(\x08\x42\x02\x18\x01\x12 \n\x14includeAnalogSamples\x18\x04 \x01(\x08\x42\x02\x18\x01\x12!\n\x15includeDigitalSamples\x18\x05 \x01(\x08\x42\x02\x18\x01\x12\x1e\n\x16includeAnalogWaveforms\x18\x06 \x01(\x08\x12\x1f\n\x17includeDigitalWaveforms\x18\x07 \x01(\x08\x12\\\n\x13simulationInterface\x18\x08 \x01(\x0b\x32?.qm.grpc.frontend.ExecutionRequest.Simulate.SimulationInterface\x12 \n\x18\x65xtraProcessingTimeoutMs\x18\t \x01(\x05\x1a\xce\x06\n\x13SimulationInterface\x12T\n\x04none\x18\x01 \x01(\x0b\x32\x44.qm.grpc.frontend.ExecutionRequest.Simulate.SimulationInterface.NoneH\x00\x12\\\n\x08loopback\x18\x02 \x01(\x0b\x32H.qm.grpc.frontend.ExecutionRequest.Simulate.SimulationInterface.LoopbackH\x00\x12[\n\x03raw\x18\x04 \x01(\x0b\x32L.qm.grpc.frontend.ExecutionRequest.Simulate.SimulationInterface.RawInterfaceH\x00\x1a\x06\n\x04None\x1a\x99\x02\n\x08Loopback\x12i\n\x0b\x63onnections\x18\x02 \x03(\x0b\x32T.qm.grpc.frontend.ExecutionRequest.Simulate.SimulationInterface.Loopback.Connections\x12\x0f\n\x07latency\x18\x03 \x01(\r\x12\x12\n\nnoisePower\x18\x04 \x01(\x01\x1a}\n\x0b\x43onnections\x12\x16\n\x0e\x66romController\x18\x01 \x01(\t\x12\x0f\n\x07\x66romFem\x18\x05 \x01(\x05\x12\x10\n\x08\x66romPort\x18\x02 \x01(\x05\x12\x14\n\x0ctoController\x18\x03 \x01(\t\x12\r\n\x05toFem\x18\x06 \x01(\x05\x12\x0e\n\x06toPort\x18\x04 \x01(\x05\x1a\xee\x01\n\x0cRawInterface\x12m\n\x0b\x63onnections\x18\x01 \x03(\x0b\x32X.qm.grpc.frontend.ExecutionRequest.Simulate.SimulationInterface.RawInterface.Connections\x12\x12\n\nnoisePower\x18\x02 \x01(\x01\x1a[\n\x0b\x43onnections\x12\x16\n\x0e\x66romController\x18\x01 \x01(\t\x12\x0f\n\x07\x66romFem\x18\x04 \x01(\x05\x12\x10\n\x08\x66romPort\x18\x02 \x01(\x05\x12\x11\n\ttoSamples\x18\x03 \x03(\x01\x42\x11\n\x0finterface_oneofB\x10\n\x0eprogram_one_of\"\xe0\x01\n\x11\x45xecutionResponse\x12\n\n\x02ok\x18\x01 \x01(\x08\x12\r\n\x05jobId\x18\x02 \x01(\t\x12\x33\n\x08messages\x18\x03 \x03(\x0b\x32!.qm.grpc.compiler.CompilerMessage\x12\x10\n\x08metadata\x18\x04 \x01(\t\x12:\n\tsimulated\x18\x06 \x01(\x0b\x32\'.qm.grpc.frontend.SimulatedResponsePart\x12-\n\x06\x63onfig\x18\x07 \x01(\x0b\x32\x1d.qm.grpc.qua_config.QuaConfig\"\xc1\x01\n\x15SimulatedResponsePart\x12\x32\n\ranalogOutputs\x18\x01 \x01(\x0b\x32\x17.google.protobuf.StructB\x02\x18\x01\x12\x33\n\x0e\x64igitalOutputs\x18\x02 \x01(\x0b\x32\x17.google.protobuf.StructB\x02\x18\x01\x12/\n\x0ewaveformReport\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x0e\n\x06\x65rrors\x18\x03 \x03(\t\"\x1c\n\x1aResetDataProcessingRequest\"\x1d\n\x1bResetDataProcessingResponse\"\x1c\n\x0bHaltRequest\x12\r\n\x05jobId\x18\x01 \x01(\t\"\x1a\n\x0cHaltResponse\x12\n\n\x02ok\x18\x01 \x01(\x08\"4\n\x0bPeekRequest\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\r\x12\x14\n\x0c\x63ontrollerId\x18\x02 \x01(\t\"\x1d\n\x0cPeekResponse\x12\r\n\x05value\x18\x01 \x01(\r\"C\n\x0bPokeRequest\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\r\x12\x14\n\x0c\x63ontrollerId\x18\x03 \x01(\t\"\x1a\n\x0cPokeResponse\x12\n\n\x02ok\x18\x01 \x01(\x08\"0\n\x1fGetSimulatedQuantumStateRequest\x12\r\n\x05jobId\x18\x01 \x01(\t\"m\n GetSimulatedQuantumStateResponse\x12\r\n\x05jobId\x18\x01 \x01(\t\x12\n\n\x02ok\x18\x02 \x01(\x08\x12.\n\x05state\x18\x03 \x01(\x0b\x32\x1f.qm.grpc.frontend.DensityMatrix\"Q\n\rDensityMatrix\x12\x11\n\ttimeStamp\x18\x01 \x01(\x04\x12-\n\x04\x64\x61ta\x18\x02 \x03(\x0b\x32\x1f.qm.grpc.frontend.ComplexNumber\"\'\n\rComplexNumber\x12\n\n\x02re\x18\x01 \x01(\x01\x12\n\n\x02im\x18\x02 \x01(\x01\"H\n\x1dPerformHalDebugCommandRequest\x12\x16\n\x0e\x63ontrollerName\x18\x01 \x01(\t\x12\x0f\n\x07\x63ommand\x18\x02 \x01(\t\"C\n\x1ePerformHalDebugCommandResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x10\n\x08response\x18\x02 \x01(\t2\x80\x19\n\x08\x46rontend\x12\x44\n\nGetVersion\x12\x16.google.protobuf.Empty\x1a\x1c.google.protobuf.StringValue\"\x00\x12N\n\x0bHealthCheck\x12\x16.google.protobuf.Empty\x1a%.qm.grpc.frontend.HealthCheckResponse\"\x00\x12T\n\x07\x45xecute\x12\".qm.grpc.frontend.ExecutionRequest\x1a#.qm.grpc.frontend.ExecutionResponse\"\x00\x12Y\n\nAddToQueue\x12#.qm.grpc.frontend.AddToQueueRequest\x1a$.qm.grpc.frontend.AddToQueueResponse\"\x00\x12P\n\x07\x43ompile\x12 .qm.grpc.frontend.CompileRequest\x1a!.qm.grpc.frontend.CompileResponse\"\x00\x12q\n\x12\x41\x64\x64\x43ompiledToQueue\x12+.qm.grpc.frontend.AddCompiledToQueueRequest\x1a,.qm.grpc.frontend.AddCompiledToQueueResponse\"\x00\x12\x64\n\x11RemovePendingJobs\x12 .qm.grpc.frontend.JobQueryParams\x1a+.qm.grpc.frontend.RemovePendingJobsResponse\"\x00\x12^\n\x0eGetPendingJobs\x12 .qm.grpc.frontend.JobQueryParams\x1a(.qm.grpc.frontend.GetPendingJobsResponse\"\x00\x12z\n\x15GetJobExecutionStatus\x12..qm.grpc.frontend.GetJobExecutionStatusRequest\x1a/.qm.grpc.frontend.GetJobExecutionStatusResponse\"\x00\x12z\n\x13GetJobStatusUpdates\x12..qm.grpc.frontend.GetJobExecutionStatusRequest\x1a/.qm.grpc.frontend.GetJobExecutionStatusResponse\"\x00\x30\x01\x12_\n\x0cPausedStatus\x12%.qm.grpc.frontend.PausedStatusRequest\x1a&.qm.grpc.frontend.PausedStatusResponse\"\x00\x12M\n\x06Resume\x12\x1f.qm.grpc.frontend.ResumeRequest\x1a .qm.grpc.frontend.ResumeResponse\"\x00\x12G\n\x04Halt\x12\x1d.qm.grpc.frontend.HaltRequest\x1a\x1e.qm.grpc.frontend.HaltResponse\"\x00\x12t\n\x13ResetDataProcessing\x12,.qm.grpc.frontend.ResetDataProcessingRequest\x1a-.qm.grpc.frontend.ResetDataProcessingResponse\"\x00\x12W\n\x0eValidateConfig\x12\x1d.qm.grpc.qua_config.QuaConfig\x1a$.qm.grpc.compiler.ValidationResponse\"\x00\x12\x38\n\x04Init\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.Empty\"\x00\x12Y\n\x10PerformQmRequest\x12 .qm.grpc.qm_api.HighQmApiRequest\x1a!.qm.grpc.qm_api.HighQmApiResponse\"\x00\x12R\n\x0bRequestData\x12\x1f.qm.grpc.frontend.QmDataRequest\x1a .qm.grpc.frontend.QmDataResponse\"\x00\x12\x85\x01\n\x14PullSimulatorSamples\x12\x35.qm.grpc.results_analyser.PullSimulatorSamplesRequest\x1a\x32.qm.grpc.results_analyser.SimulatorSamplesResponse\"\x00\x30\x01\x12u\n\x12OpenQuantumMachine\x12-.qm.grpc.qm_manager.OpenQuantumMachineRequest\x1a..qm.grpc.qm_manager.OpenQuantumMachineResponse\"\x00\x12x\n\x13\x43loseQuantumMachine\x12..qm.grpc.qm_manager.CloseQuantumMachineRequest\x1a/.qm.grpc.qm_manager.CloseQuantumMachineResponse\"\x00\x12r\n\x11GetQuantumMachine\x12,.qm.grpc.qm_manager.GetQuantumMachineRequest\x1a-.qm.grpc.qm_manager.GetQuantumMachineResponse\"\x00\x12\x66\n\rGetRunningJob\x12(.qm.grpc.qm_manager.GetRunningJobRequest\x1a).qm.grpc.qm_manager.GetRunningJobResponse\"\x00\x12h\n\x17ListOpenQuantumMachines\x12\x16.google.protobuf.Empty\x1a\x33.qm.grpc.qm_manager.ListOpenQuantumMachinesResponse\"\x00\x12h\n\x17\x43loseAllQuantumMachines\x12\x16.google.protobuf.Empty\x1a\x33.qm.grpc.qm_manager.CloseAllQuantumMachinesResponse\"\x00\x12V\n\x0eGetControllers\x12\x16.google.protobuf.Empty\x1a*.qm.grpc.qm_manager.GetControllersResponse\"\x00\x12_\n\x0cIsJobRunning\x12%.qm.grpc.frontend.IsJobRunningRequest\x1a&.qm.grpc.frontend.IsJobRunningResponse\"\x00\x12q\n\x12IsJobAcquiringData\x12+.qm.grpc.frontend.IsJobAcquiringDataRequest\x1a,.qm.grpc.frontend.IsJobAcquiringDataResponse\"\x00\x12W\n\x08Simulate\x12#.qm.grpc.frontend.SimulationRequest\x1a$.qm.grpc.frontend.SimulationResponse\"\x00\x12\x46\n\x12\x43learAllJobResults\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.Empty\"\x00\x12}\n\x16PerformHalDebugCommand\x12/.qm.grpc.frontend.PerformHalDebugCommandRequest\x1a\x30.qm.grpc.frontend.PerformHalDebugCommandResponse\"\x00\x12\x83\x01\n\x18GetSimulatedQuantumState\x12\x31.qm.grpc.frontend.GetSimulatedQuantumStateRequest\x1a\x32.qm.grpc.frontend.GetSimulatedQuantumStateResponse\"\x00\x42)\n\x10qm.grpc.frontendB\rFrontendProtoP\x01\xa2\x02\x03HLWb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'qm.pb.frontend_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020qm.grpc.frontendB\rFrontendProtoP\001\242\002\003HLW'
  _globals['_EXECUTIONOVERRIDES_WAVEFORMSENTRY']._loaded_options = None
  _globals['_EXECUTIONOVERRIDES_WAVEFORMSENTRY']._serialized_options = b'8\001'
  _globals['_GETPENDINGJOBSRESPONSE_PENDINGJOBSENTRY']._loaded_options = None
  _globals['_GETPENDINGJOBSRESPONSE_PENDINGJOBSENTRY']._serialized_options = b'8\001'
  _globals['_INTEROPXCONNECTION'].fields_by_name['source']._loaded_options = None
  _globals['_INTEROPXCONNECTION'].fields_by_name['source']._serialized_options = b'\030\001'
  _globals['_INTEROPXCONNECTION'].fields_by_name['target']._loaded_options = None
  _globals['_INTEROPXCONNECTION'].fields_by_name['target']._serialized_options = b'\030\001'
  _globals['_IOVALUEREQUEST'].fields_by_name['pulser_number']._loaded_options = None
  _globals['_IOVALUEREQUEST'].fields_by_name['pulser_number']._serialized_options = b'\030\001'
  _globals['_RESUMEREQUEST'].fields_by_name['pulsersBitmap']._loaded_options = None
  _globals['_RESUMEREQUEST'].fields_by_name['pulsersBitmap']._serialized_options = b'\030\001'
  _globals['_PAUSEDSTATUSRESPONSE'].fields_by_name['pulsersBitmap']._loaded_options = None
  _globals['_PAUSEDSTATUSRESPONSE'].fields_by_name['pulsersBitmap']._serialized_options = b'\030\001'
  _globals['_EXECUTIONREQUEST_SIMULATE'].fields_by_name['simulateAnalogOutputs']._loaded_options = None
  _globals['_EXECUTIONREQUEST_SIMULATE'].fields_by_name['simulateAnalogOutputs']._serialized_options = b'\030\001'
  _globals['_EXECUTIONREQUEST_SIMULATE'].fields_by_name['simulateDigitalOutputs']._loaded_options = None
  _globals['_EXECUTIONREQUEST_SIMULATE'].fields_by_name['simulateDigitalOutputs']._serialized_options = b'\030\001'
  _globals['_EXECUTIONREQUEST_SIMULATE'].fields_by_name['includeAnalogSamples']._loaded_options = None
  _globals['_EXECUTIONREQUEST_SIMULATE'].fields_by_name['includeAnalogSamples']._serialized_options = b'\030\001'
  _globals['_EXECUTIONREQUEST_SIMULATE'].fields_by_name['includeDigitalSamples']._loaded_options = None
  _globals['_EXECUTIONREQUEST_SIMULATE'].fields_by_name['includeDigitalSamples']._serialized_options = b'\030\001'
  _globals['_EXECUTIONREQUEST'].fields_by_name['lowLevelProgram']._loaded_options = None
  _globals['_EXECUTIONREQUEST'].fields_by_name['lowLevelProgram']._serialized_options = b'\030\001'
  _globals['_SIMULATEDRESPONSEPART'].fields_by_name['analogOutputs']._loaded_options = None
  _globals['_SIMULATEDRESPONSEPART'].fields_by_name['analogOutputs']._serialized_options = b'\030\001'
  _globals['_SIMULATEDRESPONSEPART'].fields_by_name['digitalOutputs']._loaded_options = None
  _globals['_SIMULATEDRESPONSEPART'].fields_by_name['digitalOutputs']._serialized_options = b'\030\001'
  _globals['_QUEUEPOSITION']._serialized_start=340
  _globals['_QUEUEPOSITION']._serialized_end=447
  _globals['_ADDTOQUEUEREQUEST']._serialized_start=450
  _globals['_ADDTOQUEUEREQUEST']._serialized_end=602
  _globals['_ADDTOQUEUERESPONSE']._serialized_start=604
  _globals['_ADDTOQUEUERESPONSE']._serialized_end=704
  _globals['_WAVEFORMOVERRIDE']._serialized_start=706
  _globals['_WAVEFORMOVERRIDE']._serialized_end=741
  _globals['_EXECUTIONOVERRIDES']._serialized_start=744
  _globals['_EXECUTIONOVERRIDES']._serialized_end=922
  _globals['_EXECUTIONOVERRIDES_WAVEFORMSENTRY']._serialized_start=838
  _globals['_EXECUTIONOVERRIDES_WAVEFORMSENTRY']._serialized_end=922
  _globals['_ADDCOMPILEDTOQUEUEREQUEST']._serialized_start=925
  _globals['_ADDCOMPILEDTOQUEUEREQUEST']._serialized_end=1119
  _globals['_ADDCOMPILEDTOQUEUERESPONSE']._serialized_start=1121
  _globals['_ADDCOMPILEDTOQUEUERESPONSE']._serialized_end=1232
  _globals['_COMPILEREQUEST']._serialized_start=1234
  _globals['_COMPILEREQUEST']._serialized_end=1327
  _globals['_COMPILERESPONSE']._serialized_start=1329
  _globals['_COMPILERESPONSE']._serialized_end=1430
  _globals['_GETJOBEXECUTIONSTATUSREQUEST']._serialized_start=1432
  _globals['_GETJOBEXECUTIONSTATUSREQUEST']._serialized_end=1503
  _globals['_GETJOBEXECUTIONSTATUSRESPONSE']._serialized_start=1505
  _globals['_GETJOBEXECUTIONSTATUSRESPONSE']._serialized_end=1590
  _globals['_JOBEXECUTIONSTATUS']._serialized_start=1593
  _globals['_JOBEXECUTIONSTATUS']._serialized_end=2969
  _globals['_JOBEXECUTIONSTATUS_UNKNOWN']._serialized_start=2143
  _globals['_JOBEXECUTIONSTATUS_UNKNOWN']._serialized_end=2152
  _globals['_JOBEXECUTIONSTATUS_PENDING']._serialized_start=2154
  _globals['_JOBEXECUTIONSTATUS_PENDING']._serialized_end=2252
  _globals['_JOBEXECUTIONSTATUS_RUNNING']._serialized_start=2254
  _globals['_JOBEXECUTIONSTATUS_RUNNING']._serialized_end=2376
  _globals['_JOBEXECUTIONSTATUS_COMPLETED']._serialized_start=2379
  _globals['_JOBEXECUTIONSTATUS_COMPLETED']._serialized_end=2554
  _globals['_JOBEXECUTIONSTATUS_CANCELED']._serialized_start=2556
  _globals['_JOBEXECUTIONSTATUS_CANCELED']._serialized_end=2680
  _globals['_JOBEXECUTIONSTATUS_LOADING']._serialized_start=2682
  _globals['_JOBEXECUTIONSTATUS_LOADING']._serialized_end=2755
  _globals['_JOBEXECUTIONSTATUS_ERROR']._serialized_start=2757
  _globals['_JOBEXECUTIONSTATUS_ERROR']._serialized_end=2832
  _globals['_JOBEXECUTIONSTATUS_PROCESSING']._serialized_start=2834
  _globals['_JOBEXECUTIONSTATUS_PROCESSING']._serialized_end=2959
  _globals['_GETPENDINGJOBSRESPONSE']._serialized_start=2972
  _globals['_GETPENDINGJOBSRESPONSE']._serialized_end=3174
  _globals['_GETPENDINGJOBSRESPONSE_PENDINGJOBSENTRY']._serialized_start=3078
  _globals['_GETPENDINGJOBSRESPONSE_PENDINGJOBSENTRY']._serialized_end=3174
  _globals['_JOBQUERYPARAMS']._serialized_start=3177
  _globals['_JOBQUERYPARAMS']._serialized_end=3372
  _globals['_QUERYVALUEMATCHER']._serialized_start=3374
  _globals['_QUERYVALUEMATCHER']._serialized_end=3438
  _globals['_REMOVEPENDINGJOBSRESPONSE']._serialized_start=3440
  _globals['_REMOVEPENDINGJOBSRESPONSE']._serialized_end=3497
  _globals['_SIMULATIONREQUEST']._serialized_start=3500
  _globals['_SIMULATIONREQUEST']._serialized_end=3749
  _globals['_INTEROPXADDRESS']._serialized_start=3751
  _globals['_INTEROPXADDRESS']._serialized_end=3802
  _globals['_INTEROPXTARGET']._serialized_start=3804
  _globals['_INTEROPXTARGET']._serialized_end=3883
  _globals['_INTEROPXCHANNEL']._serialized_start=3885
  _globals['_INTEROPXCHANNEL']._serialized_end=3945
  _globals['_INTEROPXCONNECTION']._serialized_start=3948
  _globals['_INTEROPXCONNECTION']._serialized_end=4505
  _globals['_INTEROPXCONNECTION_ADDRESSTOADDRESS']._serialized_start=4245
  _globals['_INTEROPXCONNECTION_ADDRESSTOADDRESS']._serialized_end=4365
  _globals['_INTEROPXCONNECTION_CHANNELTOCHANNEL']._serialized_start=4367
  _globals['_INTEROPXCONNECTION_CHANNELTOCHANNEL']._serialized_end=4487
  _globals['_SIMULATIONRESPONSE']._serialized_start=4508
  _globals['_SIMULATIONRESPONSE']._serialized_end=4831
  _globals['_ISJOBRUNNINGREQUEST']._serialized_start=4833
  _globals['_ISJOBRUNNINGREQUEST']._serialized_end=4869
  _globals['_ISJOBRUNNINGRESPONSE']._serialized_start=4871
  _globals['_ISJOBRUNNINGRESPONSE']._serialized_end=4927
  _globals['_ISJOBACQUIRINGDATAREQUEST']._serialized_start=4929
  _globals['_ISJOBACQUIRINGDATAREQUEST']._serialized_end=4971
  _globals['_ISJOBACQUIRINGDATARESPONSE']._serialized_start=4974
  _globals['_ISJOBACQUIRINGDATARESPONSE']._serialized_end=5193
  _globals['_ISJOBACQUIRINGDATARESPONSE_ACQUIRINGSTATUS']._serialized_start=5106
  _globals['_ISJOBACQUIRINGDATARESPONSE_ACQUIRINGSTATUS']._serialized_end=5193
  _globals['_QMDATAREQUEST']._serialized_start=5195
  _globals['_QMDATAREQUEST']._serialized_end=5270
  _globals['_QMDATARESPONSE']._serialized_start=5273
  _globals['_QMDATARESPONSE']._serialized_end=5554
  _globals['_QMDATARESPONSE_IOVALUERESPONSE']._serialized_start=5441
  _globals['_QMDATARESPONSE_IOVALUERESPONSE']._serialized_end=5554
  _globals['_IOVALUEREQUEST']._serialized_start=5557
  _globals['_IOVALUEREQUEST']._serialized_end=5685
  _globals['_RESUMEREQUEST']._serialized_start=5687
  _globals['_RESUMEREQUEST']._serialized_end=5744
  _globals['_RESUMERESPONSE']._serialized_start=5746
  _globals['_RESUMERESPONSE']._serialized_end=5762
  _globals['_PAUSEDSTATUSREQUEST']._serialized_start=5764
  _globals['_PAUSEDSTATUSREQUEST']._serialized_end=5800
  _globals['_PAUSEDSTATUSRESPONSE']._serialized_start=5802
  _globals['_PAUSEDSTATUSRESPONSE']._serialized_end=5881
  _globals['_PULLRESULTREQUEST']._serialized_start=5883
  _globals['_PULLRESULTREQUEST']._serialized_end=5917
  _globals['_PULLANALYSEDRESULTSREQUEST']._serialized_start=5919
  _globals['_PULLANALYSEDRESULTSREQUEST']._serialized_end=5980
  _globals['_PULLRESULTRESPONSE']._serialized_start=5983
  _globals['_PULLRESULTRESPONSE']._serialized_end=6131
  _globals['_HEALTHCHECKRESPONSE']._serialized_start=6133
  _globals['_HEALTHCHECKRESPONSE']._serialized_end=6231
  _globals['_EXECUTIONREQUEST']._serialized_start=6234
  _globals['_EXECUTIONREQUEST']._serialized_end=7750
  _globals['_EXECUTIONREQUEST_SIMULATE']._serialized_start=6522
  _globals['_EXECUTIONREQUEST_SIMULATE']._serialized_end=7732
  _globals['_EXECUTIONREQUEST_SIMULATE_SIMULATIONINTERFACE']._serialized_start=6886
  _globals['_EXECUTIONREQUEST_SIMULATE_SIMULATIONINTERFACE']._serialized_end=7732
  _globals['_EXECUTIONREQUEST_SIMULATE_SIMULATIONINTERFACE_NONE']._serialized_start=7182
  _globals['_EXECUTIONREQUEST_SIMULATE_SIMULATIONINTERFACE_NONE']._serialized_end=7188
  _globals['_EXECUTIONREQUEST_SIMULATE_SIMULATIONINTERFACE_LOOPBACK']._serialized_start=7191
  _globals['_EXECUTIONREQUEST_SIMULATE_SIMULATIONINTERFACE_LOOPBACK']._serialized_end=7472
  _globals['_EXECUTIONREQUEST_SIMULATE_SIMULATIONINTERFACE_LOOPBACK_CONNECTIONS']._serialized_start=7347
  _globals['_EXECUTIONREQUEST_SIMULATE_SIMULATIONINTERFACE_LOOPBACK_CONNECTIONS']._serialized_end=7472
  _globals['_EXECUTIONREQUEST_SIMULATE_SIMULATIONINTERFACE_RAWINTERFACE']._serialized_start=7475
  _globals['_EXECUTIONREQUEST_SIMULATE_SIMULATIONINTERFACE_RAWINTERFACE']._serialized_end=7713
  _globals['_EXECUTIONREQUEST_SIMULATE_SIMULATIONINTERFACE_RAWINTERFACE_CONNECTIONS']._serialized_start=7622
  _globals['_EXECUTIONREQUEST_SIMULATE_SIMULATIONINTERFACE_RAWINTERFACE_CONNECTIONS']._serialized_end=7713
  _globals['_EXECUTIONRESPONSE']._serialized_start=7753
  _globals['_EXECUTIONRESPONSE']._serialized_end=7977
  _globals['_SIMULATEDRESPONSEPART']._serialized_start=7980
  _globals['_SIMULATEDRESPONSEPART']._serialized_end=8173
  _globals['_RESETDATAPROCESSINGREQUEST']._serialized_start=8175
  _globals['_RESETDATAPROCESSINGREQUEST']._serialized_end=8203
  _globals['_RESETDATAPROCESSINGRESPONSE']._serialized_start=8205
  _globals['_RESETDATAPROCESSINGRESPONSE']._serialized_end=8234
  _globals['_HALTREQUEST']._serialized_start=8236
  _globals['_HALTREQUEST']._serialized_end=8264
  _globals['_HALTRESPONSE']._serialized_start=8266
  _globals['_HALTRESPONSE']._serialized_end=8292
  _globals['_PEEKREQUEST']._serialized_start=8294
  _globals['_PEEKREQUEST']._serialized_end=8346
  _globals['_PEEKRESPONSE']._serialized_start=8348
  _globals['_PEEKRESPONSE']._serialized_end=8377
  _globals['_POKEREQUEST']._serialized_start=8379
  _globals['_POKEREQUEST']._serialized_end=8446
  _globals['_POKERESPONSE']._serialized_start=8448
  _globals['_POKERESPONSE']._serialized_end=8474
  _globals['_GETSIMULATEDQUANTUMSTATEREQUEST']._serialized_start=8476
  _globals['_GETSIMULATEDQUANTUMSTATEREQUEST']._serialized_end=8524
  _globals['_GETSIMULATEDQUANTUMSTATERESPONSE']._serialized_start=8526
  _globals['_GETSIMULATEDQUANTUMSTATERESPONSE']._serialized_end=8635
  _globals['_DENSITYMATRIX']._serialized_start=8637
  _globals['_DENSITYMATRIX']._serialized_end=8718
  _globals['_COMPLEXNUMBER']._serialized_start=8720
  _globals['_COMPLEXNUMBER']._serialized_end=8759
  _globals['_PERFORMHALDEBUGCOMMANDREQUEST']._serialized_start=8761
  _globals['_PERFORMHALDEBUGCOMMANDREQUEST']._serialized_end=8833
  _globals['_PERFORMHALDEBUGCOMMANDRESPONSE']._serialized_start=8835
  _globals['_PERFORMHALDEBUGCOMMANDRESPONSE']._serialized_end=8902
  _globals['_FRONTEND']._serialized_start=8905
  _globals['_FRONTEND']._serialized_end=12105
# @@protoc_insertion_point(module_scope)



import logging
import warnings

# Suppress protobuf runtime version warnings for cross-version compatibility (must be before protobuf imports)
warnings.filterwarnings("ignore", category=UserWarning, module="google.protobuf.runtime_version")

from qm.jobs.qm_job import QmJob  # noqa: E402
from qm.version import __version__  # noqa: E402
from qm.user_config import UserConfig  # noqa: E402
from qm.program import Program, _Program  # noqa: E402
from qm.logging_utils import config_loggers  # noqa: E402
from qm.jobs.pending_job import QmPendingJob  # noqa: E402
from qm.jobs.job_queue_old_api import QmQueue  # noqa: E402
from qm.quantum_machine import QuantumMachine  # noqa: E402
from qm.api.models.capabilities import QopCaps  # noqa: E402
from qm.type_hinting import DictQuaConfig, FullQuaConfig  # noqa: E402
from qm.api.models.compiler import CompilerOptionArguments  # noqa: E402
from qm.quantum_machines_manager import QuantumMachinesManager  # noqa: E402
from qm.serialization.generate_qua_script import generate_qua_script  # noqa: E402
from qm.simulate import (  # noqa: E402
    InterOpxAddress,
    InterOpxChannel,
    InterOpxPairing,
    SimulationConfig,
    SimulatorSamples,
    LoopbackInterface,
    ControllerConnection,
    SimulatorControllerSamples,
)

from ._stream_results import (  # noqa: E402
    StreamsManager,
    StreamingResultFetcher,
    BaseSingleStreamFetcher,
    SingleStreamingResultFetcher,
    MultipleStreamingResultFetcher,
    SingleStreamSingleResultFetcher,
    SingleStreamMultipleResultFetcher,
)

__all__ = [
    "QuantumMachinesManager",
    "QuantumMachine",
    "QmPendingJob",
    "QmJob",
    "StreamsManager",
    "BaseSingleStreamFetcher",
    "StreamingResultFetcher",
    "SingleStreamingResultFetcher",
    "MultipleStreamingResultFetcher",
    "SingleStreamSingleResultFetcher",
    "SingleStreamMultipleResultFetcher",
    "Program",
    "CompilerOptionArguments",
    "generate_qua_script",
    "config_loggers",
    "UserConfig",
    "SimulationConfig",
    "InterOpxAddress",
    "InterOpxChannel",
    "ControllerConnection",
    "InterOpxPairing",
    "LoopbackInterface",
    "FullQuaConfig",
    "DictQuaConfig",
    "QopCaps",
    "SimulatorSamples",
    "SimulatorControllerSamples",
]

warnings.filterwarnings("default", category=DeprecationWarning, module="qm")
warnings.filterwarnings("ignore", category=DeprecationWarning, module="qm.grpc")

config = UserConfig.create_from_file()
config_loggers(config)


logger = logging.getLogger(__name__)
logger.info(f"Starting session: {config.SESSION_ID}")



from collections import defaultdict
from typing import Dict, Type, Union, Mapping, Iterator, Optional, Sequence

from google.protobuf.wrappers_pb2 import Int64Value

from qm.api.v2.base_api_v2 import BaseApiV2
from qm.exceptions import DataFetchingError
from qm.utils.protobuf_utils import which_one_of
from qm.api.models.server_details import ConnectionDetails
from qm.grpc.qm.grpc.v2 import job_api_pb2, common_types_pb2
from qm.grpc.qm.grpc.v2.job_api_pb2_grpc import JobServiceStub
from qm.api.models.jobs import JobNamedResult, JobStreamingState, JobResultItemSchema, JobNamedResultHeader


class JobResultApi(BaseApiV2[JobServiceStub]):
    def __init__(self, connection_details: ConnectionDetails, job_id: str, supports_chunk_streaming: bool):
        super().__init__(connection_details)
        self._id = job_id
        self._supports_chunk_streaming = supports_chunk_streaming

    @property
    def _stub_class(self) -> Type[JobServiceStub]:
        return JobServiceStub

    @property
    def id(self) -> str:
        return self._id

    def get_job_named_result(
        self, output_name: str, long_offset: int, limit: int, timeout: Optional[float] = None
    ) -> Iterator[JobNamedResult]:
        timeout = timeout if timeout is not None else self._timeout

        request = job_api_pb2.GetNamedResultRequest(
            job_id=self._id, output_name=output_name, long_offset=Int64Value(value=long_offset), limit=limit
        )
        if self._supports_chunk_streaming:
            results_iterator = self._run_iterator(self._stub.GetNamedResult, request, timeout=timeout)
            for result in self._group_results(results_iterator):
                yield result
        else:
            for response in self._stub.GetNamedResult(request, timeout=timeout):
                yield JobNamedResult(
                    data=response.success.data, count_of_items=response.success.count_of_items, output_name=output_name
                )

    def get_job_named_results(
        self, data_to_fetch: Mapping[str, slice], timeout: Optional[float]
    ) -> Iterator[JobNamedResult]:
        request = job_api_pb2.GetNamedResultsRequest(
            job_id=self._id,
            outputs=[
                job_api_pb2.GetNamedResultsRequest.Output(
                    output_name=output_name,
                    range=common_types_pb2.Range(
                        **{"from": Int64Value(value=s.start), "to": Int64Value(value=s.stop - 1)}
                    ),
                )
                for output_name, s in data_to_fetch.items()
            ],
        )
        timeout = timeout if timeout is not None else self._timeout
        results_iterator = self._stub.GetNamedResults(request, timeout=timeout)
        for result in self._group_results(results_iterator):
            yield result

    def _group_results(
        self, results_iterator: Iterator[job_api_pb2.GetNamedResultResponse]
    ) -> Iterator[JobNamedResult]:
        output_name_to_accumulated_bytes: Dict[str, bytes] = defaultdict(lambda: b"")
        # Note that the result name is a new attribute, so for old QOP we will receive just empty string.
        # But it is guaranteed that all the results belong to the same stream.
        for response in results_iterator:
            _, response_val = which_one_of(response, "response_oneof")
            if isinstance(response_val, job_api_pb2.GetNamedResultResponse.GetNamedResultResponseError):
                raise DataFetchingError(f"{response_val.details}")
            if not isinstance(response_val, job_api_pb2.GetNamedResultResponse.GetNamedResultResponseSuccess):
                raise ValueError(f"Unexpected response: {response}")
            name = response_val.output_name
            _, data_one_of = which_one_of(response_val, "data_oneof")
            if isinstance(data_one_of, job_api_pb2.GetNamedResultResponse.GetNamedResultResponseSuccess.DataChunk):
                output_name_to_accumulated_bytes[name] += data_one_of.data
            elif isinstance(data_one_of, job_api_pb2.GetNamedResultResponse.GetNamedResultResponseSuccess.DataSummary):
                data = output_name_to_accumulated_bytes.pop(name)
                result = JobNamedResult(data=data, count_of_items=data_one_of.count, output_name=name)
                yield result
            else:
                raise ValueError(f"Unexpected response: {response}")

    def get_named_headers(self, name_to_flat_struct: Mapping[str, bool]) -> Mapping[str, JobNamedResultHeader]:
        request = job_api_pb2.GetJobNamedResultsHeadersRequest(
            job_id=self._id,
            outputs=[
                job_api_pb2.GetJobNamedResultsHeadersRequest.Output(output_name=name, flat_format=flat_struct)
                for name, flat_struct in name_to_flat_struct.items()
            ],
        )
        headers = self._fetch_headers(request)
        return {response.output_name: self._convert_header_response_to_model(response) for response in headers}

    def _fetch_headers(
        self, request: job_api_pb2.GetJobNamedResultsHeadersRequest
    ) -> Sequence[job_api_pb2.GetJobNamedResultsHeadersResponse.GetJobNamedResultsHeadersResponseSuccess.OutputHeader]:
        response: job_api_pb2.GetJobNamedResultsHeadersResponse.GetJobNamedResultsHeadersResponseSuccess = self._run(
            self._stub.GetJobNamedResultsHeaders, request, timeout=self._timeout
        )
        return response.headers

    def get_named_header(self, output_name: str, flat_struct: bool) -> JobNamedResultHeader:
        request = job_api_pb2.GetJobNamedResultHeaderRequest(
            job_id=self._id, output_name=output_name, flat_format=flat_struct
        )
        response: job_api_pb2.GetJobNamedResultHeaderResponse.GetJobNamedResultHeaderResponseSuccess = self._run(
            self._stub.GetJobNamedResultHeader, request, timeout=self._timeout
        )
        return self._convert_header_response_to_model(response)

    @staticmethod
    def _convert_header_response_to_model(
        response: Union[
            job_api_pb2.GetJobNamedResultHeaderResponse.GetJobNamedResultHeaderResponseSuccess,
            job_api_pb2.GetJobNamedResultsHeadersResponse.GetJobNamedResultsHeadersResponseSuccess.OutputHeader,
        ],
    ) -> JobNamedResultHeader:
        return JobNamedResultHeader(
            count_so_far=response.count_so_far,
            bare_dtype=response.simple_dtype,
            shape=tuple(response.shape),
            has_dataloss=response.has_data_loss,
            has_execution_errors=False,
        )

    def get_job_result_schema(self) -> Mapping[str, JobResultItemSchema]:
        request = job_api_pb2.GetJobResultSchemaRequest(job_id=self._id)
        response: job_api_pb2.GetJobResultSchemaResponse.GetJobResultSchemaResponseSuccess = self._run(
            self._stub.GetJobResultSchema, request, timeout=self._timeout
        )
        return {
            item.name: JobResultItemSchema(
                name=item.name,
                bare_dtype=item.simple_dtype,
                shape=tuple(item.shape),
                is_single=item.is_single,
                expected_count=item.expected_count,
            )
            for item in response.items
        }

    def get_job_state(self) -> JobStreamingState:
        request = job_api_pb2.GetJobResultStateRequest(job_id=self._id)
        response: job_api_pb2.GetJobResultStateResponse.GetJobResultStateResponseSuccess = self._run(
            self._stub.GetJobResultState, request, timeout=self._timeout
        )
        return JobStreamingState(
            job_id=self._id,
            done=response.done,
            closed=response.closed,
            has_dataloss=response.has_dataloss,
        )

    def get_job_execution_status(self) -> common_types_pb2.JobExecutionStatus:
        request = job_api_pb2.JobServiceGetJobStatusRequest(job_id=self._id)
        response: job_api_pb2.JobServiceGetJobStatusResponse.JobServiceGetJobStatusResponseSuccess = self._run(
            self._stub.GetJobStatus, request, timeout=self._timeout
        )
        return response.status



from google.protobuf import wrappers_pb2 as _wrappers_pb2
from qm.grpc.qm.pb import general_messages_pb2 as _general_messages_pb2
from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class GetJobResultSchemaRequest(_message.Message):
    __slots__ = ("jobId",)
    JOBID_FIELD_NUMBER: _ClassVar[int]
    jobId: str
    def __init__(self, jobId: _Optional[str] = ...) -> None: ...

class GetJobResultSchemaResponse(_message.Message):
    __slots__ = ("items",)
    class Item(_message.Message):
        __slots__ = ("name", "simpleDType", "isSingle", "expectedCount", "shape")
        NAME_FIELD_NUMBER: _ClassVar[int]
        SIMPLEDTYPE_FIELD_NUMBER: _ClassVar[int]
        ISSINGLE_FIELD_NUMBER: _ClassVar[int]
        EXPECTEDCOUNT_FIELD_NUMBER: _ClassVar[int]
        SHAPE_FIELD_NUMBER: _ClassVar[int]
        name: str
        simpleDType: str
        isSingle: bool
        expectedCount: int
        shape: _containers.RepeatedScalarFieldContainer[int]
        def __init__(self, name: _Optional[str] = ..., simpleDType: _Optional[str] = ..., isSingle: bool = ..., expectedCount: _Optional[int] = ..., shape: _Optional[_Iterable[int]] = ...) -> None: ...
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    items: _containers.RepeatedCompositeFieldContainer[GetJobResultSchemaResponse.Item]
    def __init__(self, items: _Optional[_Iterable[_Union[GetJobResultSchemaResponse.Item, _Mapping]]] = ...) -> None: ...

class GetJobStateRequest(_message.Message):
    __slots__ = ("jobId",)
    JOBID_FIELD_NUMBER: _ClassVar[int]
    jobId: str
    def __init__(self, jobId: _Optional[str] = ...) -> None: ...

class GetJobStateResponse(_message.Message):
    __slots__ = ("done", "closed", "hasDataloss")
    DONE_FIELD_NUMBER: _ClassVar[int]
    CLOSED_FIELD_NUMBER: _ClassVar[int]
    HASDATALOSS_FIELD_NUMBER: _ClassVar[int]
    done: bool
    closed: bool
    hasDataloss: bool
    def __init__(self, done: bool = ..., closed: bool = ..., hasDataloss: bool = ...) -> None: ...

class GetJobNamedResultHeaderRequest(_message.Message):
    __slots__ = ("jobId", "outputName", "flatFormat")
    JOBID_FIELD_NUMBER: _ClassVar[int]
    OUTPUTNAME_FIELD_NUMBER: _ClassVar[int]
    FLATFORMAT_FIELD_NUMBER: _ClassVar[int]
    jobId: str
    outputName: str
    flatFormat: bool
    def __init__(self, jobId: _Optional[str] = ..., outputName: _Optional[str] = ..., flatFormat: bool = ...) -> None: ...

class GetJobNamedResultHeaderResponse(_message.Message):
    __slots__ = ("isSingle", "countSoFar", "simpleDType", "done", "closed", "hasDataloss", "shape", "hasExecutionErrors")
    ISSINGLE_FIELD_NUMBER: _ClassVar[int]
    COUNTSOFAR_FIELD_NUMBER: _ClassVar[int]
    SIMPLEDTYPE_FIELD_NUMBER: _ClassVar[int]
    DONE_FIELD_NUMBER: _ClassVar[int]
    CLOSED_FIELD_NUMBER: _ClassVar[int]
    HASDATALOSS_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    HASEXECUTIONERRORS_FIELD_NUMBER: _ClassVar[int]
    isSingle: bool
    countSoFar: int
    simpleDType: str
    done: bool
    closed: bool
    hasDataloss: bool
    shape: _containers.RepeatedScalarFieldContainer[int]
    hasExecutionErrors: _wrappers_pb2.BoolValue
    def __init__(self, isSingle: bool = ..., countSoFar: _Optional[int] = ..., simpleDType: _Optional[str] = ..., done: bool = ..., closed: bool = ..., hasDataloss: bool = ..., shape: _Optional[_Iterable[int]] = ..., hasExecutionErrors: _Optional[_Union[_wrappers_pb2.BoolValue, _Mapping]] = ...) -> None: ...

class GetJobNamedResultRequest(_message.Message):
    __slots__ = ("jobId", "outputName", "offset", "limit", "longOffset")
    JOBID_FIELD_NUMBER: _ClassVar[int]
    OUTPUTNAME_FIELD_NUMBER: _ClassVar[int]
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    LONGOFFSET_FIELD_NUMBER: _ClassVar[int]
    jobId: str
    outputName: str
    offset: int
    limit: int
    longOffset: _wrappers_pb2.Int64Value
    def __init__(self, jobId: _Optional[str] = ..., outputName: _Optional[str] = ..., offset: _Optional[int] = ..., limit: _Optional[int] = ..., longOffset: _Optional[_Union[_wrappers_pb2.Int64Value, _Mapping]] = ...) -> None: ...

class GetJobNamedResultResponse(_message.Message):
    __slots__ = ("countOfItems", "data")
    COUNTOFITEMS_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    countOfItems: int
    data: bytes
    def __init__(self, countOfItems: _Optional[int] = ..., data: _Optional[bytes] = ...) -> None: ...

class GetJobErrorsRequest(_message.Message):
    __slots__ = ("jobId",)
    JOBID_FIELD_NUMBER: _ClassVar[int]
    jobId: str
    def __init__(self, jobId: _Optional[str] = ...) -> None: ...

class GetJobErrorsResponse(_message.Message):
    __slots__ = ("errors", "jobId")
    class ExecutionErrorSeverity(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        ERROR: _ClassVar[GetJobErrorsResponse.ExecutionErrorSeverity]
        WARNING: _ClassVar[GetJobErrorsResponse.ExecutionErrorSeverity]
    ERROR: GetJobErrorsResponse.ExecutionErrorSeverity
    WARNING: GetJobErrorsResponse.ExecutionErrorSeverity
    class Error(_message.Message):
        __slots__ = ("errorCode", "errorSeverity", "message")
        ERRORCODE_FIELD_NUMBER: _ClassVar[int]
        ERRORSEVERITY_FIELD_NUMBER: _ClassVar[int]
        MESSAGE_FIELD_NUMBER: _ClassVar[int]
        errorCode: int
        errorSeverity: GetJobErrorsResponse.ExecutionErrorSeverity
        message: str
        def __init__(self, errorCode: _Optional[int] = ..., errorSeverity: _Optional[_Union[GetJobErrorsResponse.ExecutionErrorSeverity, str]] = ..., message: _Optional[str] = ...) -> None: ...
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    JOBID_FIELD_NUMBER: _ClassVar[int]
    errors: _containers.RepeatedCompositeFieldContainer[GetJobErrorsResponse.Error]
    jobId: str
    def __init__(self, errors: _Optional[_Iterable[_Union[GetJobErrorsResponse.Error, _Mapping]]] = ..., jobId: _Optional[str] = ...) -> None: ...

class PullAnalysedResultsRequest(_message.Message):
    __slots__ = ("jobFilePath", "metadata", "containsVersion")
    JOBFILEPATH_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    CONTAINSVERSION_FIELD_NUMBER: _ClassVar[int]
    jobFilePath: str
    metadata: str
    containsVersion: bool
    def __init__(self, jobFilePath: _Optional[str] = ..., metadata: _Optional[str] = ..., containsVersion: bool = ...) -> None: ...

class AnalysedResultsResponse(_message.Message):
    __slots__ = ("version", "icpResults", "streamResults", "errors")
    VERSION_FIELD_NUMBER: _ClassVar[int]
    ICPRESULTS_FIELD_NUMBER: _ClassVar[int]
    STREAMRESULTS_FIELD_NUMBER: _ClassVar[int]
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    version: str
    icpResults: _containers.RepeatedCompositeFieldContainer[IcpResultData]
    streamResults: _containers.RepeatedCompositeFieldContainer[StreamResultData]
    errors: _containers.RepeatedCompositeFieldContainer[_general_messages_pb2.ErrorMessage]
    def __init__(self, version: _Optional[str] = ..., icpResults: _Optional[_Iterable[_Union[IcpResultData, _Mapping]]] = ..., streamResults: _Optional[_Iterable[_Union[StreamResultData, _Mapping]]] = ..., errors: _Optional[_Iterable[_Union[_general_messages_pb2.ErrorMessage, _Mapping]]] = ...) -> None: ...

class PullResultOutputRequest(_message.Message):
    __slots__ = ("jobId", "outputName")
    JOBID_FIELD_NUMBER: _ClassVar[int]
    OUTPUTNAME_FIELD_NUMBER: _ClassVar[int]
    jobId: str
    outputName: str
    def __init__(self, jobId: _Optional[str] = ..., outputName: _Optional[str] = ...) -> None: ...

class PullResultOutputResponse(_message.Message):
    __slots__ = ("data", "header", "npz", "npy")
    class Header(_message.Message):
        __slots__ = ("version", "outputName", "dataLoss", "errors", "single")
        VERSION_FIELD_NUMBER: _ClassVar[int]
        OUTPUTNAME_FIELD_NUMBER: _ClassVar[int]
        DATALOSS_FIELD_NUMBER: _ClassVar[int]
        ERRORS_FIELD_NUMBER: _ClassVar[int]
        SINGLE_FIELD_NUMBER: _ClassVar[int]
        version: str
        outputName: str
        dataLoss: bool
        errors: _containers.RepeatedCompositeFieldContainer[_general_messages_pb2.ErrorMessage]
        single: bool
        def __init__(self, version: _Optional[str] = ..., outputName: _Optional[str] = ..., dataLoss: bool = ..., errors: _Optional[_Iterable[_Union[_general_messages_pb2.ErrorMessage, _Mapping]]] = ..., single: bool = ...) -> None: ...
    class Data(_message.Message):
        __slots__ = ("results",)
        RESULTS_FIELD_NUMBER: _ClassVar[int]
        results: _struct_pb2.Value
        def __init__(self, results: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...) -> None: ...
    class Npz(_message.Message):
        __slots__ = ("npz",)
        NPZ_FIELD_NUMBER: _ClassVar[int]
        npz: bytes
        def __init__(self, npz: _Optional[bytes] = ...) -> None: ...
    class Npy(_message.Message):
        __slots__ = ("name", "npy", "count")
        NAME_FIELD_NUMBER: _ClassVar[int]
        NPY_FIELD_NUMBER: _ClassVar[int]
        COUNT_FIELD_NUMBER: _ClassVar[int]
        name: str
        npy: bytes
        count: int
        def __init__(self, name: _Optional[str] = ..., npy: _Optional[bytes] = ..., count: _Optional[int] = ...) -> None: ...
    DATA_FIELD_NUMBER: _ClassVar[int]
    HEADER_FIELD_NUMBER: _ClassVar[int]
    NPZ_FIELD_NUMBER: _ClassVar[int]
    NPY_FIELD_NUMBER: _ClassVar[int]
    data: PullResultOutputResponse.Data
    header: PullResultOutputResponse.Header
    npz: PullResultOutputResponse.Npz
    npy: PullResultOutputResponse.Npy
    def __init__(self, data: _Optional[_Union[PullResultOutputResponse.Data, _Mapping]] = ..., header: _Optional[_Union[PullResultOutputResponse.Header, _Mapping]] = ..., npz: _Optional[_Union[PullResultOutputResponse.Npz, _Mapping]] = ..., npy: _Optional[_Union[PullResultOutputResponse.Npy, _Mapping]] = ...) -> None: ...

class BinaryWrapper(_message.Message):
    __slots__ = ("data",)
    DATA_FIELD_NUMBER: _ClassVar[int]
    data: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, data: _Optional[_Iterable[bytes]] = ...) -> None: ...

class PullFileResultRequest(_message.Message):
    __slots__ = ("jobId", "metadata", "asNpz")
    JOBID_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    ASNPZ_FIELD_NUMBER: _ClassVar[int]
    jobId: str
    metadata: str
    asNpz: bool
    def __init__(self, jobId: _Optional[str] = ..., metadata: _Optional[str] = ..., asNpz: bool = ...) -> None: ...

class FileResultResponse(_message.Message):
    __slots__ = ("jobId", "controllerName", "group", "npz", "version", "errors")
    JOBID_FIELD_NUMBER: _ClassVar[int]
    CONTROLLERNAME_FIELD_NUMBER: _ClassVar[int]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    NPZ_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    jobId: str
    controllerName: str
    group: str
    npz: bytes
    version: str
    errors: _containers.RepeatedCompositeFieldContainer[_general_messages_pb2.ErrorMessage]
    def __init__(self, jobId: _Optional[str] = ..., controllerName: _Optional[str] = ..., group: _Optional[str] = ..., npz: _Optional[bytes] = ..., version: _Optional[str] = ..., errors: _Optional[_Iterable[_Union[_general_messages_pb2.ErrorMessage, _Mapping]]] = ...) -> None: ...

class PullSimulatorSamplesRequest(_message.Message):
    __slots__ = ("jobId", "includeAnalog", "includeDigital", "asNpz", "includeAllConnections")
    JOBID_FIELD_NUMBER: _ClassVar[int]
    INCLUDEANALOG_FIELD_NUMBER: _ClassVar[int]
    INCLUDEDIGITAL_FIELD_NUMBER: _ClassVar[int]
    ASNPZ_FIELD_NUMBER: _ClassVar[int]
    INCLUDEALLCONNECTIONS_FIELD_NUMBER: _ClassVar[int]
    jobId: str
    includeAnalog: bool
    includeDigital: bool
    asNpz: bool
    includeAllConnections: bool
    def __init__(self, jobId: _Optional[str] = ..., includeAnalog: bool = ..., includeDigital: bool = ..., asNpz: bool = ..., includeAllConnections: bool = ...) -> None: ...

class SimulatorSamplesResponse(_message.Message):
    __slots__ = ("jobId", "ok", "header", "data")
    class Header(_message.Message):
        __slots__ = ("simpleDType", "countOfItems")
        SIMPLEDTYPE_FIELD_NUMBER: _ClassVar[int]
        COUNTOFITEMS_FIELD_NUMBER: _ClassVar[int]
        simpleDType: str
        countOfItems: int
        def __init__(self, simpleDType: _Optional[str] = ..., countOfItems: _Optional[int] = ...) -> None: ...
    class Data(_message.Message):
        __slots__ = ("data",)
        DATA_FIELD_NUMBER: _ClassVar[int]
        data: bytes
        def __init__(self, data: _Optional[bytes] = ...) -> None: ...
    JOBID_FIELD_NUMBER: _ClassVar[int]
    OK_FIELD_NUMBER: _ClassVar[int]
    HEADER_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    jobId: str
    ok: bool
    header: SimulatorSamplesResponse.Header
    data: SimulatorSamplesResponse.Data
    def __init__(self, jobId: _Optional[str] = ..., ok: bool = ..., header: _Optional[_Union[SimulatorSamplesResponse.Header, _Mapping]] = ..., data: _Optional[_Union[SimulatorSamplesResponse.Data, _Mapping]] = ...) -> None: ...

class IcpResultData(_message.Message):
    __slots__ = ("name", "timestamp", "dataLoss", "intValue", "doubleValue", "booleanValue")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DATALOSS_FIELD_NUMBER: _ClassVar[int]
    INTVALUE_FIELD_NUMBER: _ClassVar[int]
    DOUBLEVALUE_FIELD_NUMBER: _ClassVar[int]
    BOOLEANVALUE_FIELD_NUMBER: _ClassVar[int]
    name: str
    timestamp: int
    dataLoss: bool
    intValue: int
    doubleValue: float
    booleanValue: bool
    def __init__(self, name: _Optional[str] = ..., timestamp: _Optional[int] = ..., dataLoss: bool = ..., intValue: _Optional[int] = ..., doubleValue: _Optional[float] = ..., booleanValue: bool = ...) -> None: ...

class StreamResultData(_message.Message):
    __slots__ = ("name", "timestamp", "dataLoss", "multipleSources", "dataSourceName", "data")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DATALOSS_FIELD_NUMBER: _ClassVar[int]
    MULTIPLESOURCES_FIELD_NUMBER: _ClassVar[int]
    DATASOURCENAME_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    name: str
    timestamp: int
    dataLoss: bool
    multipleSources: bool
    dataSourceName: str
    data: int
    def __init__(self, name: _Optional[str] = ..., timestamp: _Optional[int] = ..., dataLoss: bool = ..., multipleSources: bool = ..., dataSourceName: _Optional[str] = ..., data: _Optional[int] = ...) -> None: ...



import logging
from typing import List, Tuple, cast

from qm.grpc.qm.pb import frontend_pb2
from qm.api.models.capabilities import ServerCapabilities
from qm.simulate.interface import SimulatorInterface, SupportedConnectionTypes

logger = logging.getLogger(__name__)


class LoopbackInterface(
    SimulatorInterface[frontend_pb2.ExecutionRequest.Simulate.SimulationInterface.Loopback.Connections]
):
    """Creates a loopback interface for use in
    [qm.simulate.interface.SimulationConfig][].
    A loopback connects the output of the OPX into it's input. This can be defined
    directly using the ports or through the elements.

    Args:
        connections (list):
            List of tuples with loopback connections. Each tuple should represent physical connection between ports:

                    ``(from_controller: str, from_port: int, to_controller: str, to_port: int)``

        latency (int): The latency between the OPX outputs and its
            input.
        noisePower (float): How much noise to add to the input.

                ``(fromController: str, fromFEM: int, fromPort: int, toController: str, toFEM: int, toPort: int)``

            2. Virtual connection between elements:

                ``(fromElement: str, toElement: str, toElementInput: int)``
        latency: The latency between the OPX outputs and its input.
        noisePower: How much noise to add to the input.

    Example:
        ```python
        job = qmm.simulate(config, prog, SimulationConfig(
        duration=20000,
        # loopback from output 1 to input 2 of controller 1:
        simulation_interface=LoopbackInterface([("con1", 1, "con1", 2)])
        ```
    """

    def __init__(self, connections: List[SupportedConnectionTypes], latency: int = 24, noisePower: float = 0.0):
        super().__init__(connections, noisePower)
        self._validate_latency(latency)
        self.latency = latency

    @staticmethod
    def _validate_latency(latency: int) -> None:
        if (not isinstance(latency, int)) or latency < 0:
            raise Exception("latency must be a positive integer")

    @classmethod
    def _validate_and_standardize_single_connection(
        cls, connection: SupportedConnectionTypes, fem_number_in_simulator: int
    ) -> frontend_pb2.ExecutionRequest.Simulate.SimulationInterface.Loopback.Connections:
        if not connection:
            logger.warning("No loopback was defined, treating as no loopback.")
        if not isinstance(connection, tuple):
            raise Exception("each connection must be of type tuple")
        if len(connection) == 6:
            cls._validate_connection_type(
                connection,
                [str, int, int, str, int, int],
                "(from_controller, from_fem, from_port, to_controller, to_fem, to_port)",
            )
            tuple_6 = connection
            return frontend_pb2.ExecutionRequest.Simulate.SimulationInterface.Loopback.Connections(
                fromController=tuple_6[0],
                fromFem=tuple_6[1],
                fromPort=tuple_6[2],
                toController=tuple_6[3],
                toFem=tuple_6[4],
                toPort=tuple_6[5],
            )
        if len(connection) == 4:
            cls._validate_connection_type(
                connection, [str, int, str, int], "(from_controller, from_port, to_controller, to_port)"
            )
            tuple_4 = cast(Tuple[str, int, str, int], connection)
            return frontend_pb2.ExecutionRequest.Simulate.SimulationInterface.Loopback.Connections(
                fromController=tuple_4[0],
                fromFem=fem_number_in_simulator,
                fromPort=tuple_4[1],
                toController=tuple_4[2],
                toFem=fem_number_in_simulator,
                toPort=tuple_4[3],
            )
        if len(connection) == 3:
            cls._validate_connection_type(connection, [str, str, int], "(from_Element, to_Element, to_ElementInput)")
            tuple_3 = cast(Tuple[str, str, int], connection)
            return frontend_pb2.ExecutionRequest.Simulate.SimulationInterface.Loopback.Connections(
                fromController=tuple_3[0],
                fromFem=-1,
                fromPort=-1,
                toController=tuple_3[1],
                toFem=-1,
                toPort=tuple_3[2],
            )
        raise Exception("connection should be tuple of length 3, 4 or 6")

    def update_simulate_request(
        self, request: frontend_pb2.SimulationRequest, capabilities: ServerCapabilities
    ) -> frontend_pb2.SimulationRequest:
        if not self._raw_connections:
            request.simulate.simulationInterface.none.CopyFrom(
                getattr(frontend_pb2.ExecutionRequest.Simulate.SimulationInterface, "None")()
            )
            return request

        request.simulate.simulationInterface.loopback.CopyFrom(
            frontend_pb2.ExecutionRequest.Simulate.SimulationInterface.Loopback(
                latency=self.latency,
                noisePower=self.noisePower,
                connections=self._validate_and_standardize_connections(
                    self._raw_connections, capabilities.fem_number_in_simulator
                ),
            )
        )
        return request



from abc import ABC, abstractmethod

from qm.type_hinting import NumberT
from qm.grpc.qm.pb import inc_qua_pb2
from qm.qua._expressions import QuaVariable
from qm.qua._dsl.stream_processing.stream_processing import ResultStreamSource


class DirectStreamSourceInterface(QuaVariable[NumberT], ABC):
    @abstractmethod
    def get_stream(self) -> ResultStreamSource:
        """Get or create the stream associated with DirectStreamSource for current scope."""
        pass

    @abstractmethod
    def stream_processing(self, current_processed_streams: set[str]) -> None:
        """Process and save the stream associated with DirectStreamSource.

        Mutates current_processed_streams in place, adding the names of streams processed.
        """
        pass

    @property
    @abstractmethod
    def stream_name(self) -> str:
        """Get the name of the stream associated with DirectStreamSource."""
        pass

    @abstractmethod
    def save(self) -> None:
        """Generate the save statement for the stream associated with DirectStreamSource."""
        pass

    @abstractmethod
    def get_save_statement(self) -> inc_qua_pb2.QuaProgram.AnyStatement:
        """Get the save statement for the stream associated with DirectStreamSource."""
        pass



from qm.program._dict_to_pb_converter.main_config_converter import DictToQuaConfigConverter
from qm.program._dict_to_pb_converter.converters.integration_weights_converter import build_iw_sample

__all__ = [
    "DictToQuaConfigConverter",
    "build_iw_sample",
]



import warnings
from collections.abc import Sequence
from typing import Any, Union, Literal, Optional, cast, get_args, overload

from qm.type_hinting import NumberT
from qm.exceptions import QmQuaException
from qm.qua._dsl._type_hints import OneOrMore
from qm.qua._dsl.streams.common import StreamEndpoints
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.qua._dsl.variable_handling import _standardize_value_and_size
from qm.qua._dsl.streams.external_streams import QuaStreamDirection, _declare_opnic_stream
from qm.qua._expressions import StructT, QuaArrayInputStream, QuaVariableInputStream, QuaExternalIncomingStream


def _declare_client_input_stream(
    t: type[NumberT],
    name: str,
    value: Optional[OneOrMore[Union[int, bool, float]]] = None,
    size: Optional[int] = None,
) -> Union[QuaVariableInputStream[NumberT], QuaArrayInputStream[NumberT]]:
    """Declare a QUA variable or a QUA vector to be used as an input stream from the job to the QUA program.

    Declaration is performed by declaring a python variable with the return value of this function.

    Declaration is similar to the normal QUA variable declaration. See [qm.qua.declare][] for available
    parameters.

    See [Input streams](../../Guides/features.md#input-streams) for more information.

    -- Available from QOP 2.0 --

    Example:
        ```python
        tau = declare_input_stream("client", "tau", int)
        ...
        receive_from_stream(tau)
        play('operation', 'element', duration=tau)
        ```
    """

    scope = scopes_manager.program_scope
    var = f"input_stream_{name}"

    if var in scope.declared_input_streams:
        raise QmQuaException(f"Input stream with name '{name}' already declared")

    params = _standardize_value_and_size(value, size)

    scope.add_input_stream_declaration(var)
    result = params.create_input_stream(var, t)

    scope.add_var_declaration(result.declaration_statement)

    return result


def _declare_any_input_stream(
    source: StreamEndpoints,
    stream_id: Union[int, str],
    dtype: Union[type[NumberT], type[StructT]],
    *,
    size: Optional[int] = None,
) -> Union[QuaExternalIncomingStream[StructT], QuaVariableInputStream[NumberT], QuaArrayInputStream[NumberT]]:
    if source == "client":
        # Dynamically extract valid NumberT types from TypeVar constraints
        number_types = NumberT.__constraints__  # type: ignore[misc]
        if not issubclass(dtype, tuple(number_types)):
            raise QmQuaException(
                f"Client input streams require a NumberT type (one of {', '.join(t.__name__ for t in number_types)}), "
                f"got {dtype.__name__}"
            )
        return _declare_client_input_stream(cast(type[NumberT], dtype), str(stream_id), size=size)

    elif source == "opnic":
        if size is not None and size != 1:
            raise QmQuaException("Opnic input streams currently do not support the 'size' parameter.")
        if not isinstance(stream_id, int):
            raise QmQuaException("stream_id must be an integer when declaring an opnic stream")
        return cast(
            QuaExternalIncomingStream[StructT],
            _declare_opnic_stream(cast(type[StructT], dtype), stream_id, QuaStreamDirection.INCOMING),
        )


@overload
def declare_input_stream(
    source: Literal["client"], stream_id: Union[int, str], dtype: type[NumberT]
) -> QuaVariableInputStream[NumberT]:
    """
    Declare a client input stream that carries scalar values.

    The client pushes values into the stream with [qm.jobs.base_job.QmBaseJob.push_to_input_stream][],
    and the QUA program consumes them with [qm.qua.receive_from_stream][] or
    [qm.qua.advance_input_stream][].

    Args:
        source (Literal["client"]): The endpoint type. Must be ``"client"``.
        stream_id (Union[int, str]): A unique identifier for the stream.
        dtype (type[NumberT]): The scalar type carried by the stream, such as ``int``, ``fixed``,
            ``float``, or ``bool``.
    """
    ...


@overload
def declare_input_stream(
    source: Literal["client"], stream_id: Union[int, str], dtype: type[NumberT], *, size: int
) -> QuaArrayInputStream[NumberT]:
    """
    Declare a client input stream that carries arrays.

    The client pushes arrays into the stream with [qm.jobs.base_job.QmBaseJob.push_to_input_stream][],
    and the QUA program consumes them with [qm.qua.receive_from_stream][] or
    [qm.qua.advance_input_stream][].

    Args:
        source (Literal["client"]): The endpoint type. Must be ``"client"``.
        stream_id (Union[int, str]): A unique identifier for the stream.
        dtype (type[NumberT]): The element type carried by the stream, such as ``int``, ``fixed``,
            ``float``, or ``bool``.
        size (int): The array length. Any positive integer, including ``1``, declares array data.
    """
    ...


@overload
def declare_input_stream(
    source: Literal["opnic"], stream_id: int, dtype: type[StructT]
) -> QuaExternalIncomingStream[StructT]:
    """
    Declare an incoming OPNIC packet stream.

    OPNIC streams carry packets defined by a type decorated with [qm.qua.qua_struct][].
    The QUA program receives packets with [qm.qua.receive_from_stream][] into a struct instance
    declared with [qm.qua.declare_struct][].

    Args:
        source (Literal["opnic"]): The endpoint type. Must be ``"opnic"``.
        stream_id (int): A unique integer stream ID.
        dtype (type[StructT]): The packet type carried by the stream. This must be a type decorated
            with [qm.qua.qua_struct][].
    """
    ...


# ======================================================
#  Regular input stream declaration, deprecated signature
# ======================================================


@overload
def declare_input_stream(t: type[NumberT], name: str) -> QuaVariableInputStream[NumberT]:
    ...


@overload
def declare_input_stream(t: type[NumberT], name: str, value: Literal[None], size: int) -> QuaArrayInputStream[NumberT]:
    ...


@overload
def declare_input_stream(t: type[NumberT], name: str, *, size: int) -> QuaArrayInputStream[NumberT]:
    ...


@overload
def declare_input_stream(
    t: type[NumberT], name: str, value: Union[int, bool, float]
) -> QuaVariableInputStream[NumberT]:
    ...


@overload
def declare_input_stream(
    t: type[NumberT], name: str, value: Sequence[Union[int, bool, float]]
) -> QuaArrayInputStream[NumberT]:
    ...


def declare_input_stream(
    *args: Any, **kwargs: Any
) -> Union[QuaExternalIncomingStream[StructT], QuaVariableInputStream[NumberT], QuaArrayInputStream[NumberT]]:
    """
    Declare an input stream.

    This function supports client input streams, incoming OPNIC packet streams, and a deprecated
    client-only signature.

    Args:
        args (Any): Accepts one of the following positional argument sets:
            * ``"client"``, ``stream_id``, ``dtype`` to declare a client input stream.
            * ``"client"``, ``stream_id``, ``dtype`` with ``size=...`` to declare a client input stream
              that carries arrays.
            * ``"opnic"``, ``stream_id``, ``PacketType`` to declare an incoming OPNIC packet stream.
            * ``dtype``, ``name`` for the deprecated client-only signature.

            Where:
            * ``stream_id``: A unique identifier for the stream. Client streams accept integers or strings.
              OPNIC streams require an integer stream ID.
            * ``dtype``: The scalar type carried by a client stream, such as ``int``, ``fixed``, ``float``,
              or ``bool``.
            * ``PacketType``: A type decorated with [qm.qua.qua_struct][] that defines the packet schema
              carried by an OPNIC stream.

        kwargs (Any): Supports the following keyword arguments:

            * ``size`` (int): Client streams only. If omitted, the stream carries scalar values.
              If provided, the stream carries arrays of that length. OPNIC streams do not support ``size``.
            * Deprecated client-only signature keyword arguments such as ``value`` and ``size``, with the
              same behavior as [qm.qua.declare][].

    Returns:
        The declared input stream.

    Examples:
        ```python
        tau = declare_input_stream("client", "tau_input", int)
        truth_table = declare_input_stream("client", "truth_table", bool, size=10)

        @qua_struct
        class Packet:
            data: QuaArray[int, 1]

        incoming_packet = declare_input_stream("opnic", 1, Packet)
        ```

    """
    if isinstance(args[0], str) and args[0] in get_args(StreamEndpoints):
        return _declare_any_input_stream(*args, **kwargs)
    else:
        warnings.warn(
            "The signature `declare_input_stream(t, name , ...)`, which implicitly treats the "
            "stream as a Client stream, is deprecated and will be removed in a future release. "
            'Please use `declare_input_stream("client", stream_id=name, dtype=t, ...) instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        return _declare_client_input_stream(*args, **kwargs)



import abc
from typing import Union, TypeVar, Optional

from qm._loc import _get_loc
from qm.type_hinting import Number
from qm.grpc.qm.pb import inc_qua_pb2
from qm.exceptions import QmQuaException
from qm.utils.protobuf_utils import which_one_of
from qm.qua._expressions import QuaVariable, QuaArrayCell, QuaArrayVariable
from qm.qua._dsl.stream_processing.direct_stream_processing_interface import DirectStreamSourceInterface


class AnalogTimeDivision(metaclass=abc.ABCMeta):
    def __init__(self, loc: str, samples_per_chunk: int):
        self.loc = loc
        self.samples_per_chunk = samples_per_chunk

    @property
    @abc.abstractmethod
    def unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogProcessTarget.TimeDivision:
        pass


class SlicedAnalogTimeDivision(AnalogTimeDivision):
    @property
    def unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogProcessTarget.TimeDivision:
        return inc_qua_pb2.QuaProgram.AnalogProcessTarget.TimeDivision(
            sliced=inc_qua_pb2.QuaProgram.AnalogTimeDivision.Sliced(samplesPerChunk=self.samples_per_chunk)
        )


class AccumulatedAnalogTimeDivision(AnalogTimeDivision):
    @property
    def unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogProcessTarget.TimeDivision:
        return inc_qua_pb2.QuaProgram.AnalogProcessTarget.TimeDivision(
            accumulated=inc_qua_pb2.QuaProgram.AnalogTimeDivision.Accumulated(samplesPerChunk=self.samples_per_chunk)
        )


class MovingWindowAnalogTimeDivision(AnalogTimeDivision):
    def __init__(self, loc: str, samples_per_chunk: int, chunks_per_window: int):
        super().__init__(loc, samples_per_chunk)
        self.chunks_per_window = chunks_per_window

    @property
    def unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogProcessTarget.TimeDivision:
        return inc_qua_pb2.QuaProgram.AnalogProcessTarget.TimeDivision(
            movingWindow=inc_qua_pb2.QuaProgram.AnalogTimeDivision.MovingWindow(
                samplesPerChunk=self.samples_per_chunk,
                chunksPerWindow=self.chunks_per_window,
            )
        )


class AnalogProcessTarget(metaclass=abc.ABCMeta):
    def __init__(self, loc: str):
        self.loc = loc

    @property
    @abc.abstractmethod
    def unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogProcessTarget:
        pass

    @abc.abstractmethod
    def save_stream_if_needed(self) -> None:
        pass


class ScalarProcessTarget(AnalogProcessTarget):
    def __init__(self, loc: str, target: Union[QuaVariable[float], QuaArrayCell[float]]):
        super().__init__(loc)
        self.target = target

    @property
    def unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogProcessTarget:
        target_exp = self.target.unwrapped
        if not isinstance(target_exp, inc_qua_pb2.QuaProgram.AnyScalarExpression):
            raise QmQuaException(f"Unknown type - {type(target_exp)}")

        target_type, found = which_one_of(target_exp, "expression_oneof")
        if isinstance(found, inc_qua_pb2.QuaProgram.VarRefExpression):
            target = inc_qua_pb2.QuaProgram.AnalogProcessTarget.ScalarProcessTarget(variable=found)
        elif isinstance(found, inc_qua_pb2.QuaProgram.ArrayCellRefExpression):
            target = inc_qua_pb2.QuaProgram.AnalogProcessTarget.ScalarProcessTarget(arrayCell=found)
        else:
            raise QmQuaException(f"Unknown target type - {target_type}")
        return inc_qua_pb2.QuaProgram.AnalogProcessTarget(scalarProcess=target)

    def save_stream_if_needed(self) -> None:
        if isinstance(self.target, DirectStreamSourceInterface):
            self.target.save()


class VectorProcessTarget(AnalogProcessTarget):
    def __init__(self, loc: str, target: QuaArrayVariable[float], time_division: AnalogTimeDivision):
        super().__init__(loc)
        self.time_division = time_division
        self.target = target

    @property
    def unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogProcessTarget:
        target_exp = self.target.unwrapped
        target = inc_qua_pb2.QuaProgram.AnalogProcessTarget.VectorProcessTarget(
            array=target_exp,
            timeDivision=self.time_division.unwrapped,
        )
        return inc_qua_pb2.QuaProgram.AnalogProcessTarget(vectorProcess=target)

    def save_stream_if_needed(self) -> None:
        if isinstance(self.target, DirectStreamSourceInterface):
            self.target.save()


class MeasureProcessAbstract(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        self.loc = _get_loc()

    @property
    @abc.abstractmethod
    def unwrapped(self) -> inc_qua_pb2.QuaProgram.MeasureProcess:
        pass

    @abc.abstractmethod
    def save_stream_if_needed(self) -> None:
        pass


class AnalogMeasureProcess(MeasureProcessAbstract, metaclass=abc.ABCMeta):
    def __init__(self, target: AnalogProcessTarget):
        super().__init__()
        self.target = target

    @property
    def unwrapped(self) -> inc_qua_pb2.QuaProgram.MeasureProcess:
        return inc_qua_pb2.QuaProgram.MeasureProcess(analog=self._analog_unwrapped)

    @property
    @abc.abstractmethod
    def _analog_unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogMeasureProcess:
        pass

    def save_stream_if_needed(self) -> None:
        self.target.save_stream_if_needed()


class BasicIntegration(AnalogMeasureProcess):
    def __init__(self, element_output: str, iw: str, target: AnalogProcessTarget):
        super().__init__(target)
        self.element_output = element_output
        self.iw = iw

    @property
    @abc.abstractmethod
    def _analog_unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogMeasureProcess:
        pass


class BareIntegration(BasicIntegration):
    @property
    def _analog_unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogMeasureProcess:
        return inc_qua_pb2.QuaProgram.AnalogMeasureProcess(
            loc=self.loc,
            bareIntegration=inc_qua_pb2.QuaProgram.AnalogMeasureProcess.BareIntegration(
                integration=inc_qua_pb2.QuaProgram.IntegrationWeightReference(name=self.iw),
                elementOutput=self.element_output,
                target=self.target.unwrapped,
            ),
        )


class DemodIntegration(BasicIntegration):
    @property
    def _analog_unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogMeasureProcess:
        return inc_qua_pb2.QuaProgram.AnalogMeasureProcess(
            loc=self.loc,
            demodIntegration=inc_qua_pb2.QuaProgram.AnalogMeasureProcess.DemodIntegration(
                integration=inc_qua_pb2.QuaProgram.IntegrationWeightReference(name=self.iw),
                elementOutput=self.element_output,
                target=self.target.unwrapped,
            ),
        )


class DualMeasureProcess(AnalogMeasureProcess, metaclass=abc.ABCMeta):
    def __init__(
        self,
        element_output1: str,
        element_output2: str,
        iw1: str,
        iw2: str,
        target: AnalogProcessTarget,
    ):
        super().__init__(target)
        self.element_output1 = element_output1
        self.element_output2 = element_output2
        self.iw1 = iw1
        self.iw2 = iw2


class DualDemodIntegration(DualMeasureProcess):
    @property
    def _analog_unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogMeasureProcess:
        return inc_qua_pb2.QuaProgram.AnalogMeasureProcess(
            loc=self.loc,
            dualDemodIntegration=inc_qua_pb2.QuaProgram.AnalogMeasureProcess.DualDemodIntegration(
                integration1=inc_qua_pb2.QuaProgram.IntegrationWeightReference(name=self.iw1),
                integration2=inc_qua_pb2.QuaProgram.IntegrationWeightReference(name=self.iw2),
                elementOutput1=self.element_output1,
                elementOutput2=self.element_output2,
                target=self.target.unwrapped,
            ),
        )


T = TypeVar(
    "T",
    bound=Union[
        inc_qua_pb2.QuaProgram.AnalogMeasureProcess.RawTimeTagging,
        inc_qua_pb2.QuaProgram.AnalogMeasureProcess.HighResTimeTagging,
        inc_qua_pb2.QuaProgram.DigitalMeasureProcess.RawTimeTagging,
    ],
)


def _add_target_len(time_tagging: T, target_len: Optional[QuaVariable[int]]) -> T:
    if target_len is not None:
        time_tagging.targetLen.CopyFrom(target_len.unwrapped.variable)
    return time_tagging


class TimeTaggingMeasurementProcess(MeasureProcessAbstract, metaclass=abc.ABCMeta):
    def __init__(
        self,
        element_output: str,
        target: QuaArrayVariable[int],
        targetLen: Optional[QuaVariable[int]],
        max_time: Number,
    ):
        super().__init__()
        self.time_tagging_target = target
        self.element_output = element_output
        self.targetLen = targetLen
        self.max_time = max_time

    @property
    def unwrapped_target(self) -> inc_qua_pb2.QuaProgram.ArrayVarRefExpression:
        a = self.time_tagging_target.unwrapped
        assert isinstance(a, inc_qua_pb2.QuaProgram.ArrayVarRefExpression)
        return a

    @property
    def unwrapped(self) -> inc_qua_pb2.QuaProgram.MeasureProcess:
        return inc_qua_pb2.QuaProgram.MeasureProcess(analog=self._analog_unwrapped)

    @property
    @abc.abstractmethod
    def _analog_unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogMeasureProcess:
        pass

    def save_stream_if_needed(self) -> None:
        pass


class RawTimeTagging(TimeTaggingMeasurementProcess):
    @property
    def _analog_unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogMeasureProcess:
        time_tagging = inc_qua_pb2.QuaProgram.AnalogMeasureProcess.RawTimeTagging(
            maxTime=int(self.max_time),
            elementOutput=self.element_output,
            target=self.unwrapped_target,
        )
        time_tagging = _add_target_len(time_tagging, self.targetLen)
        return inc_qua_pb2.QuaProgram.AnalogMeasureProcess(loc=self.loc, rawTimeTagging=time_tagging)


class HighResTimeTagging(TimeTaggingMeasurementProcess):
    @property
    def _analog_unwrapped(self) -> inc_qua_pb2.QuaProgram.AnalogMeasureProcess:
        time_tagging = inc_qua_pb2.QuaProgram.AnalogMeasureProcess.HighResTimeTagging(
            maxTime=int(self.max_time),
            elementOutput=self.element_output,
            target=self.unwrapped_target,
        )
        time_tagging = _add_target_len(time_tagging, self.targetLen)
        return inc_qua_pb2.QuaProgram.AnalogMeasureProcess(loc=self.loc, highResTimeTagging=time_tagging)



from collections import defaultdict
from typing import Dict, List, Union, Optional, Sequence

from google.protobuf.struct_pb2 import Struct

from qm.api.v2.job_api import JobApi
from qm.grpc.qm.grpc.v2 import job_api_pb2
from qm.waveform_report import WaveformReport
from qm.utils.config_utils import get_fem_config
from qm.utils.protobuf_utils import which_one_of
from qm.jobs.simulated_job import extract_struct_value
from qm.api.models.server_details import ConnectionDetails
from qm.grpc.qm.pb import frontend_pb2, inc_qua_config_pb2
from qm.api.v2.job_api.job_api import JobApiWithDeprecations
from qm.exceptions import QopResponseError, QMSimulationError
from qm.api.models.capabilities import QopCaps, ServerCapabilities
from qm._stream_results import SimulatorSamples, SimulatorControllerSamples


class SimulatedJobApi(JobApi):
    def __init__(
        self,
        connection_details: ConnectionDetails,
        job_id: str,
        simulated_response: Optional[frontend_pb2.SimulatedResponsePart],
        capabilities: ServerCapabilities,
    ) -> None:
        super().__init__(connection_details, job_id, capabilities)
        self._waveform_report = None

        # In QOP 3.2 and earlier, the waveform_report is included inside the simulated_response, which is provided
        # during the creation of the SimulatedJobApi. This if statement is checking this case.
        # TODO: Remove the support for this flow in QOP 3.6
        if not capabilities.supports(QopCaps.waveform_report_endpoint) and simulated_response:
            self._waveform_report = self._build_waveform_report(simulated_response.waveformReport)

    def _build_waveform_report(self, raw_waveform_report: Struct) -> WaveformReport:
        return WaveformReport.from_dict(extract_struct_value(raw_waveform_report), self.id)

    def _get_raw_waveform_report_from_api(self) -> Struct:
        request = job_api_pb2.GetWaveformReportRequest(job_id=self.id)
        try:
            success: job_api_pb2.GetWaveformReportResponse.GetWaveformReportSuccess = self._run(
                self._stub.GetWaveformReport, request, timeout=self._timeout
            )
        except QopResponseError as e:
            raise QMSimulationError("Error while getting waveform report from API. Error: " + str(e))

        return success.waveformReport

    def get_simulated_waveform_report(self) -> WaveformReport:
        if self._waveform_report is None:
            raw_waveform_report = self._get_raw_waveform_report_from_api()
            self._waveform_report = self._build_waveform_report(raw_waveform_report)

        return self._waveform_report

    def _pull_simulator_samples(
        self, include_analog: bool, include_digital: bool
    ) -> Dict[str, List[job_api_pb2.PullSamplesResponse.PullSamplesResponseSuccess]]:
        request = job_api_pb2.PullSamplesRequest(
            job_id=self._id, include_analog=include_analog, include_digital=include_digital
        )
        bare_results = defaultdict(list)
        for result in self._stub.PullSamples(request, timeout=self._timeout):
            _, response = which_one_of(result, "response_oneof")
            if isinstance(response, job_api_pb2.PullSamplesResponse.PullSamplesResponseSuccess):
                bare_results[response.controller].append(response)
            else:
                raise QMSimulationError("Error while pulling samples")
        return dict(bare_results)

    def get_simulated_samples(self, include_analog: bool = True, include_digital: bool = True) -> SimulatorSamples:
        config = self._get_pb_config()

        results_by_controller = self._pull_simulator_samples(include_analog, include_digital)
        controller_to_samples = {}
        for controller, responses in results_by_controller.items():
            analog: Dict[str, Sequence[Union[float, complex]]] = {}
            digital = {}
            analog_sampling_rate = {}
            for response in responses:
                fem_config = get_fem_config(
                    config,
                    inc_qua_config_pb2.QuaConfig.AdcPortReference(
                        controller=controller, fem=response.fem_id, number=response.port_id
                    ),
                )
                key = f"{response.fem_id}-{response.port_id}"
                if response.mode == job_api_pb2.PullSamplesResponse.PullSamplesResponseSuccess.Mode.ANALOG:
                    if isinstance(fem_config, inc_qua_config_pb2.QuaConfig.ControllerDec):
                        analog_sampling_rate[key] = 1e9
                    elif isinstance(fem_config, inc_qua_config_pb2.QuaConfig.OctoDacFemDec):
                        analog_sampling_rate[key] = 2e9
                    elif isinstance(fem_config, inc_qua_config_pb2.QuaConfig.MicrowaveFemDec):
                        sampling_rate = fem_config.analogOutputs[response.port_id].samplingRate
                        assert (
                            sampling_rate is not None
                        )  # Mypy thinks it can be None, but it can't really (sampling_rate has a default value)
                        analog_sampling_rate[key + "-1"] = sampling_rate
                        analog_sampling_rate[key + "-2"] = sampling_rate
                    else:
                        raise QMSimulationError(f"Unknown FEM type: {fem_config}")

                    samples = response.double_data
                    _, data = which_one_of(samples, "output")
                    if isinstance(data, job_api_pb2.PullSamplesResponse.PullSamplesResponseSuccess.mw):
                        analog[f"{key}-{data.ducId}"] = [x + 1j * y for x, y in zip(data.I, data.Q)]
                    elif isinstance(data, job_api_pb2.PullSamplesResponse.PullSamplesResponseSuccess.lf):
                        analog[key] = data.items
                else:
                    digital[key] = response.boolean_data.data[0].item
            controller_to_samples[controller] = SimulatorControllerSamples(
                analog, digital, analog_sampling_rate=analog_sampling_rate
            )
        return SimulatorSamples(controller_to_samples)

    def plot_waveform_report_with_simulated_samples(self) -> None:
        self.get_simulated_waveform_report().create_plot(self.get_simulated_samples())

    def plot_waveform_report_without_samples(self) -> None:
        self.get_simulated_waveform_report().create_plot()


class SimulatedJobApiWithDeprecations(JobApiWithDeprecations, SimulatedJobApi):
    pass



from qm._loc import _get_loc
from qm.grpc.qm.pb import inc_qua_pb2
from qm.exceptions import QmQuaException
from qm.qua._expressions import Scalar, to_scalar_pb_expression
from qm.qua._scope_management.scopes_manager import scopes_manager


def update_frequency(
    element: str,
    new_frequency: Scalar[int],
    units: str = "Hz",
    keep_phase: bool = False,
) -> None:
    """Dynamically update the frequency of the oscillator associated with a given `element`.

    This changes the frequency from the value defined in the quantum machine configuration.

    The behavior of the phase (continuous vs. coherent) is controlled by the ``keep_phase`` parameter and
    is discussed in [the documentation](../../Introduction/qua_overview.md#frequency-and-phase-transformations).

    Args:
        element (str): The element associated with the oscillator whose
            frequency will be changed
        new_frequency (int): The new frequency value to set in units set
            by ``units`` parameter. In steps of 1.
        units (str): units of new frequency. Useful when sub-Hz
            precision is required. Allowed units are "Hz", "mHz", "uHz",
            "nHz", "pHz"
        keep_phase (bool): Determine whether phase will be continuous
            through the change (if `True`) or it will be coherent,
            only the frequency will change (if `False`).

    Example:
        ```python
        with program() as prog:
            update_frequency("q1", 4e6) # will set the frequency to 4 MHz

            ### Example for sub-Hz resolution
            update_frequency("q1", 100.7) # will set the frequency to 100 Hz (due to casting to int)
            update_frequency("q1", 100700, units='mHz') # will set the frequency to 100.7 Hz
        ```
    """
    try:
        units_enum = getattr(inc_qua_pb2.QuaProgram.UpdateFrequencyStatement.Units, units)
    except KeyError:
        raise QmQuaException(f'unknown units "{units}"')

    loc = _get_loc()
    statement = inc_qua_pb2.QuaProgram.UpdateFrequencyStatement(
        loc=loc,
        qe=inc_qua_pb2.QuaProgram.QuantumElementReference(name=element, loc=loc),
        units=units_enum,
        keepPhase=keep_phase,
        value=to_scalar_pb_expression(new_frequency),
    )
    scopes_manager.append_statement(inc_qua_pb2.QuaProgram.AnyStatement(updateFrequency=statement))


def update_correction(
    element: str,
    c00: Scalar[float],
    c01: Scalar[float],
    c10: Scalar[float],
    c11: Scalar[float],
) -> None:
    """Updates the correction matrix used to overcome IQ imbalances of the IQ mixer for the next pulses
    played on the element

    Note:

        Make sure to update the correction after you called [`update_frequency`][qm.qua.update_frequency]

    Note:

        Up to QOP 3.3, calling ``update_correction`` will also reset the frame of the oscillator associated with the element.

    Args:
        element (str): The element associated with the oscillator whose
            correction matrix will change
        c00 (Union[float,QUA variable of type real]): The top left
            matrix element
        c01 (Union[float,QUA variable of type real]): The top right
            matrix element
        c10 (Union[float,QUA variable of type real]): The bottom left
            matrix element
        c11 (Union[float,QUA variable of type real]): The bottom right
            matrix element

    Example:
        ```python
        with program() as prog:
            update_correction("q1", 1.0, 0.5, 0.5, 1.0)
        ```
    """
    loc = _get_loc()
    statement = inc_qua_pb2.QuaProgram.UpdateCorrectionStatement(
        loc=loc,
        qe=inc_qua_pb2.QuaProgram.QuantumElementReference(name=element, loc=loc),
        correction=inc_qua_pb2.QuaProgram.Correction(
            c0=to_scalar_pb_expression(c00),
            c1=to_scalar_pb_expression(c01),
            c2=to_scalar_pb_expression(c10),
            c3=to_scalar_pb_expression(c11),
        ),
    )
    scopes_manager.append_statement(inc_qua_pb2.QuaProgram.AnyStatement(updateCorrection=statement))


def set_dc_offset(element: str, element_input: str, offset: Scalar[float]) -> None:
    """Set the DC offset of an element's input to the given value.
    Up to QOP 3.5, this value will remain the DC offset until changed or until the Quantum Machine is closed.
    From QOP 3.5, the DC offset will be reset to the QM's idle value at the end of each program.

    -- Available from QOP 2.0 --

    Args:
        element: The element to update its DC offset
        element_input: The desired input of the element, can be 'single'
            for a 'singleInput' element or 'I' or 'Q' for a 'mixInputs'
            element
        offset: The offset to set
    """
    loc = _get_loc()
    statement = inc_qua_pb2.QuaProgram.SetDcOffsetStatement(
        loc=loc,
        qe=inc_qua_pb2.QuaProgram.QuantumElementReference(name=element, loc=loc),
        qeInputReference=element_input,
        offset=to_scalar_pb_expression(offset),
    )
    scopes_manager.append_statement(inc_qua_pb2.QuaProgram.AnyStatement(setDcOffset=statement))



import logging
import warnings
from dataclasses import field, dataclass
from typing import Dict, List, Type, Union, Literal, Mapping, Iterable, Optional, MutableSequence

from qm.octave import QmOctaveConfig
from qm.grpc.qm.grpc.v2 import qmm_api_pb2
from qm.api.v2.base_api_v2 import BaseApiV2
from qm.octave.octave_manager import OctaveManager
from qm.utils.protobuf_utils import proto_map_to_dict
from qm.api.v2.qm_api_old import QmApiWithDeprecations
from qm.type_hinting.config_types import FullQuaConfig
from qm.api.models.server_details import ConnectionDetails
from qm.grpc.qm.grpc.v2.qmm_api_pb2_grpc import QmmServiceStub
from qm.api.v2.job_api.simulated_job_api import SimulatedJobApi
from qm.api.models.capabilities import QopCaps, ServerCapabilities
from qm.grpc.qm.pb import inc_qua_pb2, frontend_pb2, inc_qua_config_pb2
from qm.api.v2.qm_api import QmApi, _log_messages, handle_simulation_error
from qm.exceptions import OpenQmException, QopResponseError, JobNotFoundException
from qm.api.v2.job_api.job_api import JobApi, JobData, JobStatus, JobApiWithDeprecations, transfer_statuses_to_enum

logger = logging.getLogger(__name__)


@dataclass
class VersionResponse:
    gateway: str
    controllers: Dict[str, str]


ControllerTypes = Literal["OPX", "OPX1000"]
FemTypes = Literal["LF", "MW"]
FEM_TYPES_MAPPING: Dict[int, FemTypes] = {1: "LF", 2: "MW"}


@dataclass
class ControllerBase:
    name: str

    @property
    def controller_type(self) -> ControllerTypes:
        raise NotImplementedError


@dataclass
class Controller(ControllerBase):
    _temperature: Optional[float] = field(repr=False)

    @property
    def temperature(self) -> Optional[float]:
        if self._temperature is None:
            logger.warning(
                "Temperature is not available for this controller, "
                "Either because it uses an old gateway or its state is not known."
            )
        return self._temperature

    @property
    def controller_type(self) -> ControllerTypes:
        return "OPX"


@dataclass
class ControllerOPX1000(ControllerBase):
    hostname: str
    fems: Dict[int, FemTypes]
    _temperatures: Optional[Dict[str, float]]

    @property
    def controller_type(self) -> ControllerTypes:
        return "OPX1000"

    @property
    def temperatures(self) -> Optional[Mapping[str, float]]:
        if self._temperatures is None:
            logger.warning(
                f"Temperatures are not supported in this QOP {QopCaps.device_temperatures.from_qop_version} version"
            )
        return self._temperatures


class QmmApi(BaseApiV2[QmmServiceStub]):
    QM_CLASS = QmApi
    JOB_CLASS = JobApi

    def __init__(
        self,
        connection_details: ConnectionDetails,
        capabilities: ServerCapabilities,
        octave_config: Optional[QmOctaveConfig],
        octave_manager: OctaveManager,
    ) -> None:
        super().__init__(connection_details)
        self._caps = capabilities
        self._octave_config = octave_config
        self._octave_manager = octave_manager

    @property
    def _stub_class(self) -> Type[QmmServiceStub]:
        return QmmServiceStub

    def get_qm(self, qm_id: str, _pb_config: Optional[inc_qua_config_pb2.QuaConfig] = None) -> QmApi:
        # todo - remove _pb_config when octave config is in the GW
        return self.QM_CLASS(
            connection_details=self.connection_details,
            qm_id=qm_id,
            capabilities=self._caps,
            octave_config=self._octave_config,
            octave_manager=self._octave_manager,
            pb_config=_pb_config,
        )

    def get_job(self, job_id: str) -> JobApi:
        jobs_data = self.get_jobs(job_ids=[job_id])
        if not jobs_data:
            raise JobNotFoundException(job_id)
        else:
            job_data = jobs_data[0]
            if self._caps.supports(QopCaps.waveform_report_endpoint) and job_data.is_simulation:
                return self._get_simulated_job(job_id)

            return self.JOB_CLASS(self.connection_details, job_id, capabilities=self._caps)

    def get_jobs(
        self,
        qm_ids: Iterable[str] = tuple(),
        job_ids: Iterable[str] = tuple(),
        user_ids: Iterable[str] = tuple(),
        description: str = "",
        status: Union[JobStatus, Iterable[JobStatus]] = tuple(),
    ) -> List[JobData]:
        query_params = qmm_api_pb2.JobsQueryParams(
            quantum_machine_ids=list(qm_ids),
            job_ids=list(job_ids),
            user_ids=list(user_ids),
            description=description,
            status=transfer_statuses_to_enum(status),
        )
        request = qmm_api_pb2.GetJobsRequest(query=query_params)
        response: qmm_api_pb2.GetJobsSuccess = self._run(self._stub.GetJobs, request, timeout=self._timeout)
        return [JobData.from_grpc(j) for j in response.jobs]

    def open_qm(self, config: inc_qua_config_pb2.QuaConfig, close_other_machines: Optional[bool] = None) -> QmApi:
        if close_other_machines is None:
            warnings.warn(
                "close_other_machines is not set, as from 2.0.0 default will be False, now setting to True. Please set it explicitly to remove this message and keep the wanted behavior in future versions.",
                DeprecationWarning,
            )
            close_other_machines = True
        request = qmm_api_pb2.OpenQuantumMachineRequest(
            config=config,
            close_mode=(
                qmm_api_pb2.OpenQuantumMachineRequest.CloseMode.CLOSE_MODE_IF_NEEDED
                if close_other_machines
                else qmm_api_pb2.OpenQuantumMachineRequest.CloseMode.CLOSE_MODE_UNSPECIFIED
            ),
        )

        try:
            response: qmm_api_pb2.OpenQuantumMachineResponse.OpenQuantumMachineResponseSuccess = self._run(
                self._stub.OpenQuantumMachine, request, timeout=self._timeout
            )

        except QopResponseError as e:
            error = e.error
            open_qm_exception = OpenQmException(error.config_validation_errors, error.physical_validation_errors)

            for formatted_error in (
                open_qm_exception.physical_validation_formatted_errors
                + open_qm_exception.config_validation_formatted_errors
            ):
                logger.error(formatted_error)

            raise open_qm_exception

        for warning in response.open_qm_warnings:
            logger.warning(f"Open QM ended with warning {warning.code}: {warning.message}")

        logger.info("Opened quantum machine with id: " + response.quantum_machine_id)

        return self.get_qm(response.quantum_machine_id, _pb_config=config)

    def perform_healthcheck(self) -> None:
        logger.info("Performing health check")
        response: qmm_api_pb2.HealthCheckResponse.HealthCheckResponseSuccess = self._run(
            self._stub.HealthCheck, qmm_api_pb2.HealthCheckRequest(), timeout=self._timeout
        )
        msg = "Cluster healthcheck completed successfully."
        if response.details:
            msg += " Details:"
            for k, v in response.details.items():
                msg += f"\n  {k}: {v}"
        logger.info(msg)

    def get_version(self) -> VersionResponse:
        response: qmm_api_pb2.GetVersionResponse.GetVersionResponseSuccess = self._run(
            self._stub.GetVersion, qmm_api_pb2.GetVersionRequest(), timeout=self._timeout
        )
        return VersionResponse(
            gateway=response.gateway,
            controllers={k: v for k, v in response.controllers.items()},
        )

    def get_controllers(self) -> Mapping[str, ControllerBase]:
        response: qmm_api_pb2.GetControllersResponse.GetControllersResponseSuccess = self._run(
            self._stub.GetControllers, qmm_api_pb2.GetControllersRequest(), timeout=self._timeout
        )
        to_return = {}
        correction_offset = 0 if self._caps.supports(QopCaps.opx1000_fems_return_1_based) else 1
        for name, value in response.control_devices.items():
            if value.controller_type == 1:
                to_return[name] = ControllerOPX1000(
                    name=name,
                    hostname=value.hostname,
                    fems={
                        int(i) + correction_offset: FEM_TYPES_MAPPING[f.type]
                        for i, f in value.fems.items()
                        if f.type > 0
                    },
                    _temperatures=proto_map_to_dict(value.temperatures)
                    if self._caps.supports(QopCaps.device_temperatures)
                    else None,
                )
            else:
                raise NotImplementedError(f"Controller type {value.controller_type} is not supported.")
        return to_return

    def reset_data_processing(self) -> None:
        request = qmm_api_pb2.QmmServiceResetDataProcessingRequest()
        self._run(self._stub.ResetDataProcessing, request, timeout=self._timeout)

    def validate_qua_config(self, config: FullQuaConfig) -> None:
        pass

    def _get_simulated_job(
        self, job_id: str, simulated: Union[frontend_pb2.SimulatedResponsePart, None] = None
    ) -> SimulatedJobApi:
        return self.QM_CLASS.SIMULATED_JOB_CLASS(
            self.connection_details, job_id, simulated_response=simulated, capabilities=self._caps
        )

    def simulate(
        self,
        config: inc_qua_config_pb2.QuaConfig,
        program: inc_qua_pb2.QuaProgram,
        simulate: frontend_pb2.ExecutionRequest.Simulate,
        controller_connections: MutableSequence[frontend_pb2.InterOpxConnection],
    ) -> SimulatedJobApi:
        request = qmm_api_pb2.QmmServiceSimulateRequest(
            config=config,
            high_level_program=program,
            simulate=simulate,
            controller_connections=controller_connections,
        )
        logger.info("Simulating program.")

        with handle_simulation_error():
            response: qmm_api_pb2.SimulationSuccess = self._run(self._stub.Simulate, request, timeout=self._timeout)

        _log_messages(response.messages)

        return self._get_simulated_job(response.job_id, response.simulated)

    def list_open_qms(self) -> List[str]:
        request = qmm_api_pb2.ListOpenQuantumMachinesRequest()
        response: qmm_api_pb2.ListOpenQuantumMachinesResponse.ListOpenQuantumMachinesResponseSuccess = self._run(
            self._stub.ListOpenQuantumMachines, request, timeout=self._timeout
        )
        return list(response.machine_ids)

    def close_all_qms(self) -> None:
        request = qmm_api_pb2.CloseAllQuantumMachinesRequest()
        self._run(self._stub.CloseAllQuantumMachines, request, timeout=self._timeout)

    def clear_all_job_results(self) -> None:
        request = qmm_api_pb2.ClearAllJobResultsRequest()
        self._run(self._stub.ClearAllJobResults, request, timeout=self._timeout)


class QmmApiWithDeprecations(QmmApi):
    QM_CLASS = QmApiWithDeprecations
    JOB_CLASS = JobApiWithDeprecations



from typing import Type, Tuple, Union, Optional, overload

from qm.grpc.qm.grpc.v2 import job_api_pb2
from qm.api.v2.job_api.generic_apis import ElementGenericApi
from qm.grpc.qm.pb import inc_qm_api_pb2, inc_qua_config_pb2

InputConfigType = Union[
    inc_qua_config_pb2.QuaConfig.SingleInput,
    inc_qua_config_pb2.QuaConfig.MixInputs,
    inc_qua_config_pb2.QuaConfig.SingleInputCollection,
    inc_qua_config_pb2.QuaConfig.MultipleInputs,
    inc_qua_config_pb2.QuaConfig.MicrowaveInputPortReference,
    None,
]


class UnknownElementType(ValueError):
    pass


class ElementInputApi(ElementGenericApi):
    pass


class NoInputApi(ElementInputApi):
    pass


class MultipleInputsApi(ElementInputApi):
    pass


class SingleInputCollectionApi(ElementInputApi):
    pass


class MixInputsApi(ElementInputApi):
    def set_correction(self, correction: Tuple[float, float, float, float]) -> None:
        matrix = inc_qm_api_pb2.Matrix(v00=correction[0], v01=correction[1], v10=correction[2], v11=correction[3])
        request = job_api_pb2.SetMatrixCorrectionRequest(job_id=self._id, qe=self._element_id, correction=matrix)
        self._run(self._stub.SetMatrixCorrection, request, timeout=self._timeout)

    def get_correction(self) -> Tuple[float, float, float, float]:
        request = job_api_pb2.GetMatrixCorrectionRequest(job_id=self._id, qe=self._element_id)
        correction: inc_qm_api_pb2.Matrix = self._run(
            self._stub.GetMatrixCorrection, request, timeout=self._timeout
        ).correction
        return correction.v00, correction.v01, correction.v10, correction.v11

    def set_dc_offsets(self, i_offset: Optional[float] = None, q_offset: Optional[float] = None) -> None:
        request = job_api_pb2.SetOutputDcOffsetRequest(
            job_id=self._id, qe=self._element_id, mix_inputs=job_api_pb2.MixInputsDcOffset(I=i_offset, Q=q_offset)
        )
        self._run(self._stub.SetOutputDcOffset, request, timeout=self._timeout)

    def get_dc_offsets(self) -> Tuple[float, float]:
        request = job_api_pb2.GetOutputDcOffsetRequest(job_id=self._id, qe=self._element_id)
        response: job_api_pb2.GetOutputDcOffsetResponse.GetOutputDcOffsetResponseSuccess = self._run(
            self._stub.GetOutputDcOffset, request, timeout=self._timeout
        )
        i, q = response.mix_inputs.I, response.mix_inputs.Q
        if i is None or q is None:
            raise ValueError("Mixed DC offsets are not set")
        return i, q


class SingleInputApi(ElementInputApi):
    def set_dc_offset(self, offset: float) -> None:
        request = job_api_pb2.SetOutputDcOffsetRequest(
            job_id=self._id, qe=self._element_id, single_input=job_api_pb2.SingleInputDcOffset(offset=offset)
        )
        self._run(self._stub.SetOutputDcOffset, request, timeout=self._timeout)

    def get_dc_offset(self) -> float:
        request = job_api_pb2.GetOutputDcOffsetRequest(job_id=self._id, qe=self._element_id)
        response: job_api_pb2.GetOutputDcOffsetResponse.GetOutputDcOffsetResponseSuccess = self._run(
            self._stub.GetOutputDcOffset, request, timeout=self._timeout
        )
        return response.single_input.offset


class MwInputApi(ElementInputApi):
    def set_converter_frequency(self, frequency_hz: float, set_also_output: bool = True) -> None:
        request = job_api_pb2.SetOscillatorFrequencyRequest(
            job_id=self._id,
            qe=self._element_id,
            new_frequency_hz=frequency_hz,
            update_component=(
                job_api_pb2.SetOscillatorFrequencyRequest.UpdateComponentSelection.both
                if set_also_output
                else job_api_pb2.SetOscillatorFrequencyRequest.UpdateComponentSelection.upconverter
            ),
        )
        self._run(self._stub.SetOscillatorFrequency, request, timeout=self._timeout)


@overload
def create_element_input_class(input_config: inc_qua_config_pb2.QuaConfig.SingleInput) -> Type[SingleInputApi]:
    pass


@overload
def create_element_input_class(input_config: inc_qua_config_pb2.QuaConfig.MixInputs) -> Type[MixInputsApi]:
    pass


@overload
def create_element_input_class(
    input_config: inc_qua_config_pb2.QuaConfig.SingleInputCollection,
) -> Type[SingleInputCollectionApi]:
    pass


@overload
def create_element_input_class(input_config: inc_qua_config_pb2.QuaConfig.MultipleInputs) -> Type[MultipleInputsApi]:
    pass


@overload
def create_element_input_class(
    input_config: inc_qua_config_pb2.QuaConfig.MicrowaveInputPortReference,
) -> Type[MwInputApi]:
    pass


@overload
def create_element_input_class(input_config: None) -> Type[NoInputApi]:
    pass


def create_element_input_class(input_config: InputConfigType) -> Type[ElementInputApi]:
    if isinstance(input_config, inc_qua_config_pb2.QuaConfig.SingleInput):
        return SingleInputApi
    if isinstance(input_config, inc_qua_config_pb2.QuaConfig.MixInputs):
        return MixInputsApi
    if isinstance(input_config, inc_qua_config_pb2.QuaConfig.SingleInputCollection):
        return SingleInputCollectionApi
    if isinstance(input_config, inc_qua_config_pb2.QuaConfig.MultipleInputs):
        return MultipleInputsApi
    if isinstance(input_config, inc_qua_config_pb2.QuaConfig.MicrowaveInputPortReference):
        return MwInputApi
    if input_config is None:
        return NoInputApi
    raise UnknownElementType(f"Unknown input type {type(input_config)}")



from typing import Any, List, Optional

from qm.qua.extensions.qua_iterators.qua_iterators_base import IterableBase
from qm.qua.extensions.qua_iterators.qua_iterators_types import QuaNamedTuple, MultiIteratorType, IteratorContentTypes


class QuaProduct:
    """
    Combine iterables into nested loops, similarly to Python's
    ``itertools.product``.

    ``QuaProduct`` expands a sequence of iterable helpers into nested loops.
    The first iterable becomes the outermost loop and the last iterable becomes
    the innermost loop, so reordering the list changes the loop nesting order.

    A product can mix QUA iterables and native iterables. The yielded value is
    a named tuple whose field names match the iterable names.

    Note:
        Use ``QuaProduct`` as the outer combinator. Do not pass a
        ``QuaProduct`` instance into another iterable helper such as
        [QuaZip][qm.qua.extensions.qua_iterators.QuaZip].

    Example:
        ```python
        with program() as prog:
            for args in QuaProduct(
                [
                    QuaIterableRange("shot", 100),
                    NativeIterable("element", ["q1", "q2"]),
                    QuaZip(
                        [
                            QuaIterable("amp", [0.2, 0.5, 0.8]),
                            QuaIterable("tau", [16, 32, 64]),
                        ],
                        name="drive",
                    ),
                ]
            ):
                play("x90" * amp(args.drive.amp), args.element)
                wait(args.drive.tau)
        ```
    """

    def __init__(self, iterables: List[IterableBase[Any]]):
        self._iterables = iterables
        self._iterable_names = [itr.name for itr in self._iterables]

    @property
    def iterables(self) -> tuple[IterableBase[Any], ...]:
        return tuple(self._iterables)

    def _traverse_iterables(
        self, iterables_id: int = 0, yielded_itr: Optional[List[IteratorContentTypes]] = None
    ) -> MultiIteratorType:
        """
        Recursively traverse the product iterables from first to last.

        Args:
            iterables_id: Index of the iterable currently being traversed.
            yielded_itr: Values yielded by the iterables visited so far.
        """
        if not yielded_itr:
            yielded_itr = []
        if iterables_id == len(self._iterables):
            yield QuaNamedTuple(self._iterable_names, yielded_itr)
        else:
            for curr_itr in self._iterables[iterables_id]:
                if iterables_id == len(yielded_itr):
                    yielded_itr.append(curr_itr)
                else:
                    yielded_itr[iterables_id] = curr_itr
                yield from self._traverse_iterables(iterables_id + 1, yielded_itr)

    def __iter__(self) -> MultiIteratorType:
        yield from self._traverse_iterables()



from qm.octave_sdk.connectivity.connectivity_util import RFInputRFSource
from qm.octave_sdk.octave import IFMode, Octave, ClockType, RFOutputMode, ClockFrequency
from qm.octave_sdk.connectivity.connectivity import OctaveOutput, OctaveLOSource, RFInputLOSource

__all__ = [
    "RFInputRFSource",
    "OctaveLOSource",
    "RFInputLOSource",
    "OctaveOutput",
    "Octave",
    "ClockType",
    "ClockFrequency",
    "IFMode",
    "RFOutputMode",
]
